"""
A script to invoke utilities from either uvx or from local installs.

I got tired of installing mypy/pyright in a million venvs. I got around this
for pyright using a central install and running in an activated venv. You can
do likewise with mypy, by passing `--python-executable`. I got the bright idea
to use `uvx` (pipx originally) to manage mypy and pyright. But when working
from my own machine, I'd already have the type checkers installed centrally.
This script automates running these tools. It does the following:

* Optionally can set the `specification` (i.e., "mypy==1.2.3...", etc)
* You can set the specifications from a `requirements.txt` file. So you can use
  tools like `pip-compile` to manage the versions.
* Makes the `--python-executable` and `--pythonpath` flags to mypy and pyright
  the same. Defaults to using `sys.executable` from the python running this
* Also sets `--python-version` and `--pythonversion` in mypy and pyright
* For other tools, just run them from using uvx (same as calling uvx directly)
"""

from __future__ import annotations

import logging
import os
import re
import shlex
import subprocess
import sys
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, cast

from packaging.requirements import Requirement

if TYPE_CHECKING:
    import argparse
    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from types import ModuleType

# Optional imports
colorlog: ModuleType | None
try:
    import colorlog
except ImportError:
    colorlog = None


# * Logger --------------------------------------------------------------------
# Taken from nox.logger (https://github.com/wntrblm/nox/tree/main/nox/logger.py)

SUCCESS = 25
OUTPUT = logging.DEBUG - 1


def _get_format(color: bool, add_timestamp: bool) -> str:
    if color and colorlog:
        if add_timestamp:
            return "%(cyan)s%(name)s > [%(asctime)s] %(log_color)s%(message)s"
        return "%(cyan)s%(name)s > %(log_color)s%(message)s"

    if add_timestamp:
        return "%(name)s > [%(asctime)s] %(message)s"

    return "%(name)s > %(message)s"


class UvxRunFormatter(logging.Formatter):
    """Custom formatter."""

    def __init__(self, add_timestamp: bool = False) -> None:
        super().__init__(fmt=_get_format(color=False, add_timestamp=add_timestamp))
        self._simple_fmt = logging.Formatter("%(message)s")

    def format(self, record: Any) -> str:
        if record.levelname == "OUTPUT":
            return self._simple_fmt.format(record)
        return super().format(record)


if colorlog:

    class UvxRunColoredFormatter(colorlog.ColoredFormatter):  # type: ignore[misc,name-defined]
        """Colored formatter."""

        def __init__(
            self,
            datefmt: Any = None,
            style: Any = None,
            log_colors: Any = None,
            reset: bool = True,
            secondary_log_colors: Any = None,
            add_timestamp: bool = False,
        ) -> None:
            super().__init__(  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
                fmt=_get_format(color=True, add_timestamp=add_timestamp),
                datefmt=datefmt,
                style=style,
                log_colors=log_colors,
                reset=reset,
                secondary_log_colors=secondary_log_colors,
            )
            self._simple_fmt = logging.Formatter("%(message)s")

        def format(self, record: Any) -> str:
            if record.levelname == "OUTPUT":
                return self._simple_fmt.format(record)
            return super().format(record)  # type: ignore[no-any-return]


class LoggerWithSuccessAndOutput(logging.getLoggerClass()):  # type: ignore[misc]
    """Custom Logger."""

    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)
        logging.addLevelName(SUCCESS, "SUCCESS")
        logging.addLevelName(OUTPUT, "OUTPUT")

    def success(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(SUCCESS):  # pragma: no cover
            self._log(SUCCESS, msg, args, **kwargs)

    def output(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(OUTPUT):  # pragma: no cover
            self._log(OUTPUT, msg, args, **kwargs)


logging.setLoggerClass(LoggerWithSuccessAndOutput)
logger = cast(LoggerWithSuccessAndOutput, logging.getLogger("uvxrun"))


def _get_formatter(color: bool, add_timestamp: bool) -> logging.Formatter:
    if color and colorlog:
        return UvxRunColoredFormatter(
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "blue",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
                "SUCCESS": "green",
            },
            style="%",
            secondary_log_colors=None,
            add_timestamp=add_timestamp,
        )
    return UvxRunFormatter(add_timestamp=add_timestamp)


def setup_logging(
    color: bool, verbosity: int = 0, add_timestamp: bool = False
) -> None:  # pragma: no cover
    """Setup logging.

    Args:
        color (bool): If true, the output will be colored using
            colorlog. Otherwise, it will be plaintext.
    """
    root_logger = logging.getLogger()

    level_number = max(0, logging.WARNING - 10 * verbosity)
    root_logger.setLevel(level_number)

    handler = logging.StreamHandler()

    handler.setFormatter(_get_formatter(color, add_timestamp))
    root_logger.addHandler(handler)

    # Silence noisy loggers
    logging.getLogger("sh").setLevel(logging.WARNING)


# * Utilities -----------------------------------------------------------------
@lru_cache
def _comment_re() -> re.Pattern[str]:
    return re.compile(r"(^|\s+)#.*$")


def _parse_requirements(requirements: Path) -> Iterator[Requirement]:
    comment_re = _comment_re()

    def ignore_comments(lines: Iterable[str]) -> Iterator[str]:
        """Strips comments and filter empty lines."""
        for line in lines:
            if line_formatted := comment_re.sub("", line).strip():
                yield line_formatted

    with requirements.open() as f:
        yield from map(Requirement, ignore_comments(f))


@lru_cache
def _get_python_version(python_version: str | None) -> str:
    if python_version is None:
        return "{}.{}".format(*sys.version_info[:2])
    return python_version


@lru_cache
def _get_python_executable(python_executable: str | None) -> str:
    if python_executable is None:
        return sys.executable
    return python_executable


def _iter_specs(specs: Iterable[str | Requirement]) -> Iterator[Requirement]:
    for spec in specs:
        if isinstance(spec, str):
            yield Requirement(spec)
        else:
            yield spec


# * Specs ---------------------------------------------------------------------
class Specifications:
    """Working with specifications."""

    def __init__(self, specs: dict[str, Requirement]) -> None:
        self.specs = specs

    def assign(
        self, specs: str | Requirement | Iterable[str | Requirement]
    ) -> Specifications:
        if isinstance(specs, (str, Requirement)):
            specs = [specs]

        return type(self)(
            specs=dict(self.specs, **{spec.name: spec for spec in _iter_specs(specs)})
        )

    def get(self, index: str, default: Requirement | None = None) -> Requirement | None:
        return self.specs.get(index, default)

    @classmethod
    def combine(cls, *specs: Specifications) -> Specifications:
        new_specs: dict[str, Requirement] = {}
        for spec in specs:
            new_specs.update(spec.specs)
        return cls(new_specs)

    @classmethod
    def from_requirements(
        cls,
        *specs: Requirement,
        requirements: str | Path | Iterable[str | Path] | None = None,
    ) -> Specifications:
        if requirements is None:
            requirements = []
        elif isinstance(requirements, (str, Path)):
            requirements = [requirements]

        specs_dict: dict[str, Requirement] = {
            req.name: req
            for requirement in requirements
            for req in _parse_requirements(Path(requirement))
        }

        # update specs
        specs_dict.update({spec.name: spec for spec in _iter_specs(specs)})

        return cls(specs_dict)


# * Dummy session -------------------------------------------------------------


class SessionInterfaceTemplate(Protocol):
    """Generic session."""

    def run(
        self,
        *args: str | os.PathLike[str],
        env: Mapping[str, str | None] | None = None,
        include_outer_env: bool = True,
    ) -> Any | None: ...

    def log(self, *args: Any, **kwargs: Any) -> None: ...

    def warn(self, *args: Any, **kwargs: Any) -> None: ...

    def debug(self, *args: Any, **kwargs: Any) -> None: ...


def _clean_env(env: Mapping[str, str | None] | None = None) -> dict[str, str] | None:
    if env is None:
        return None
    return {k: v for k, v in env.items() if v is not None}


class Session:
    """Basic session"""

    def log(self, *args: Any, **kwargs: Any) -> None:
        """Alias to self.info"""
        self.info(*args, **kwargs)

    @staticmethod
    def info(*args: Any, **kwargs: Any) -> None:
        """Output log info."""
        logger.info(*args, **kwargs)

    @staticmethod
    def warn(*args: Any, **kwargs: Any) -> None:
        """Output log.warning"""
        logger.warning(*args, **kwargs)

    @staticmethod
    def debug(*args: Any, **kwargs: Any) -> None:
        """Debug."""
        logger.debug(*args, **kwargs)

    def run(
        self,
        *args: str | os.PathLike[str],
        env: Mapping[str, str | None] | None = None,
        include_outer_env: bool = True,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Actually do the run."""
        cleaned_args = [os.fsdecode(arg) for arg in args]

        full_cmd = shlex.join(cleaned_args)

        self.info("Running %s", full_cmd)
        r = subprocess.run(cleaned_args, check=False, env=_clean_env(env))  # pyright: ignore[reportUnknownVariableType]

        returncode: int = r.returncode  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        if returncode != 0:
            logger.error("Command %s failed with exit code %s", full_cmd, returncode)  # pyright: ignore[reportUnknownArgumentType]
            msg = f"Returned code {returncode}"
            raise RuntimeError(msg)


# * Command -------------------------------------------------------------------
class Command:
    """Class to handle commands."""

    def __init__(
        self, name: str, args: Iterable[str], spec: Requirement | None = None
    ) -> None:
        self.name = name
        self.args = list(args)
        self.spec = spec

    @classmethod
    def from_command(cls, *commands: str | os.PathLike[str]) -> Command:
        """Create from command iterable"""
        name, *args = map(os.fsdecode, commands)
        spec = Requirement(name)
        return cls(
            name=spec.name,
            args=args,
            spec=spec if spec.specifier else None,
        )

    def assign_spec(self, spec: Requirement | None, override: bool = False) -> Command:
        """Update spec from specs dict."""
        if spec and (override or not self.spec):
            return type(self)(name=self.name, args=self.args, spec=spec)
        return self

    def _get_python_flags(
        self,
        python_executable: str | None,
        python_version: str | None,
    ) -> list[str]:
        if self.name == "mypy":
            return [
                f"--python-executable={_get_python_executable(python_executable)}",
                f"--python-version={_get_python_version(python_version)}",
            ]
        if self.name == "pyright":
            return [
                f"--pythonpath={_get_python_executable(python_executable)}",
                f"--pythonversion={_get_python_version(python_version)}",
            ]

        return []

    @staticmethod
    def _get_uvx_flags(verbosity: int) -> list[str]:
        if verbosity > 0:
            return ["-v"]
        if verbosity < 0:
            return ["-q"]
        return []

    def command(
        self,
        session: SessionInterfaceTemplate,
        python_executable: str | None,
        python_version: str | None,
        files: str | Iterable[str] | None,
        verbosity: int = 0,
    ) -> list[str]:
        """Create command list."""
        if files is None:
            files = []
        elif isinstance(files, str):
            files = [files]

        commands: list[str]
        session.log("Using uvx %s", self.spec or self.name)
        commands = [
            "uvx",
            *self._get_uvx_flags(verbosity=verbosity),
            *([f"--from={self.spec}"] if self.spec else []),
            self.name,
        ]

        return [
            *commands,
            *self._get_python_flags(
                python_executable=python_executable,
                python_version=python_version,
            ),
            *self.args,
            *files,
        ]

    def run(
        self,
        session: SessionInterfaceTemplate,
        python_executable: str | None,
        python_version: str | None,
        files: Iterable[str] | None = None,
        dry: bool = False,
        verbosity: int = 0,
        env: Mapping[str, str] | None = None,
        include_outer_env: bool = True,
        **kwargs: Any,
    ) -> Any:
        """Run the command"""
        command = self.command(
            session=session,
            python_executable=python_executable,
            python_version=python_version,
            files=files,
            verbosity=verbosity,
        )

        if dry:
            session.log("Would run with commands: %s", command)
            return 0

        return session.run(
            *command, env=env, include_outer_env=include_outer_env, **kwargs
        )


def run(
    *args: str | os.PathLike[str],
    session: SessionInterfaceTemplate | None = None,
    specs: Specifications | None = None,
    extra_specs: str | Requirement | Iterable[str | Requirement] | None = None,
    python_version: str | None = None,
    python_executable: str | None = None,
    verbosity: int = -2,
    env: Mapping[str, str] | None = None,
    files: Iterable[str] | None = None,
    include_outer_env: bool = True,
    **kwargs: Any,
) -> Any:
    """Run command"""
    command = Command.from_command(*args)

    if specs is None:
        specs = Specifications({})

    if extra_specs is not None:
        specs = specs.assign(extra_specs)

    command = command.assign_spec(specs.get(command.name))

    session = session or Session()

    return command.run(
        session=session,
        python_executable=python_executable,
        python_version=python_version,
        files=files,
        verbosity=verbosity,
        env=env,
        include_outer_env=include_outer_env,
        **kwargs,
    )


# * CLI -----------------------------------------------------------------------
def _parse_args(args: Sequence[str] | None = None) -> argparse.Namespace:
    """Get parser."""
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Run executable using uvx.")
    parser.add_argument(
        "--python-executable",
        dest="python_executable",
        default=None,  # Path(sys.executable),
        type=Path,
        help="""
        Path to python executable. Defaults to ``sys.executable``. This is
        passed to `--python-executable` in mypy and `--pythonpath` in pyright.
        """,
    )
    parser.add_argument(
        "--python-version",
        dest="python_version",
        default=None,
        type=str,
        help="""
        Python version (x.y) to typecheck against. Defaults to
        ``{sys.version_info.major}.{sys.version_info.minor}``. This is passed
        to ``--python-version`` and ``--pythonversion`` in mypy and pyright.
        """,
    )
    parser.add_argument(
        "-c",
        "--command",
        dest="commands",
        default=[],
        action="append",
        type=str,
        help="""
        Checkers command. This can include extra flags to the checker. This can
        also be passed multiple times for multiple checkers. This can also
        include a Requirement specification, which overrides any specification
        from ``--requirement`` or ``--spec``. For example,
        ``--command='mypy==1.8.0 --no-incremental' -c pyright``
        """,
    )
    parser.add_argument(
        "-r",
        "--requirement",
        dest="requirements",
        default=[],
        action="append",
        type=Path,
        help="Requirements (requirements.txt) specs for checker.  Can specify multiple times.",
    )
    parser.add_argument(
        "-s",
        "--spec",
        dest="specs",
        default=[],
        action="append",
        type=Requirement,
        help="""
        Package specification. Can pass multiple times. Overrides specs read
        from ``--requirements``. For example, ``--spec 'mypy==1.2.3'``.
        """,
    )
    parser.add_argument("--dry", action="store_true", help="Do dry run")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="Set verbosity level.  Pass multiple times to up level.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="Lower verbosity level.  Pass multiple times to up level.",
    )
    parser.add_argument(
        "--fail-fast",
        dest="fail_fast",
        action="store_true",
        help="If passed, exit at first error.",
    )
    parser.add_argument(
        "files", type=str, default=[], nargs="*", help="Optional files to check."
    )

    options = parser.parse_args() if args is None else parser.parse_args(args)

    if not options.commands:
        parser.print_usage()

    return options


def main(args: Sequence[str] | None = None) -> int:
    """Main script."""
    if not (options := _parse_args(args)).commands:
        return 0

    # setup logging:
    setup_logging(
        color=True, verbosity=options.verbosity - options.quiet, add_timestamp=False
    )

    logger.info(
        "Running with python %s at %s",
        _get_python_version(None),
        _get_python_executable(None),
    )

    commands = [
        Command.from_command(*shlex.split(command)) for command in options.commands
    ]

    # specs from requirements
    specs = Specifications.from_requirements(
        *options.specs, requirements=options.requirements
    )

    # update command specs
    commands = [command.assign_spec(specs.get(command.name)) for command in commands]

    session = Session()

    s = 0

    for command in commands:
        try:
            command.run(
                session=session,
                python_executable=options.python_executable,
                python_version=options.python_version,
                dry=options.dry,
                verbosity=options.verbosity - options.quiet - 2,
                files=options.files,
            )
        except RuntimeError:  # noqa: PERF203
            s += 1
            if options.fail_fast:
                return s
    return s


if __name__ == "__main__":
    sys.exit(main())



"""
Read/use config/pyproject.toml file (:mod:`~pyproject2conda.config`)
====================================================================
"""

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from pyproject2conda.utils import (
    filename_from_template,
    get_all_pythons,
    get_default_pythons,
    get_in,
    select_pythons,
)

if TYPE_CHECKING:
    from typing import Any, Iterator, Sequence

    from ._typing_compat import Self


# * Utilities
class Config:  # noqa: PLR0904
    """Class to parse toml file with [tool.pyproject2conda] section"""

    def __init__(
        self,
        data: dict[str, Any],
        default_pythons: list[str] | None = None,
        all_pythons: list[str] | None = None,
    ) -> None:
        self.data = data
        self.default_pythons: list[str] = (
            [] if default_pythons is None else default_pythons
        )
        self.all_pythons: list[str] = [] if all_pythons is None else all_pythons

    def get_in(self, *keys: str, default: Any = None) -> Any:
        """Utility to extract from nested dict."""
        return get_in(keys=keys, nested_dict=self.data, default=default)

    @cached_property
    def overrides(self) -> list[Any]:
        """All overrides sections from `[[tool.pyproject2conda.overrides]]`"""
        out: list[dict[str, Any]] = []
        for x in self.get_in("overrides", default=[]):
            if "envs" not in x:
                msg = "must specify env in overrides"
                raise ValueError(msg)
            out.append(x)
        return out

    @property
    def envs(self) -> dict[str, Any]:
        """All environments"""
        return self.get_in("envs", default={})  # type: ignore[no-any-return]

    def _get_override(self, env: str) -> dict[str, Any]:
        out: dict[str, Any] = {}
        for override in self.overrides:
            if env in override["envs"]:
                out.update(**override)
        out.pop("envs", None)
        return out

    def _get_value(
        self,
        key: str,
        env_name: str | None = None,
        inherit: bool = True,
        as_list: bool = False,
        default: Any = None,
    ) -> Any:
        """Get a value from thing"""
        if env_name is None:
            value = self.get_in(key, default=None)

        else:
            # try to get from env definition
            if env_name not in self.data["envs"]:
                msg = f"env {env_name} not in config"
                raise ValueError(msg)

            value = self.get_in("envs", env_name, key, default=None)

            if inherit:
                # If have override, use it.
                _value = self._get_override(env_name).get(key, None)
                if _value is not None:
                    value = _value

                # finally, try to get from top level
                if value is None:
                    value = self.get_in(key, default=None)

        if value is None:
            value = default() if callable(default) else default

        if value is not None and as_list and not isinstance(value, list):
            value = [value]

        return value  # pyright: ignore[reportUnknownVariableType]

    def channels(
        self, env_name: str | None = None, inherit: bool = True
    ) -> list[str] | None:
        """Channels getter"""
        return self._get_value(  # type: ignore[no-any-return]
            key="channels", env_name=env_name, inherit=inherit, as_list=True
        )

    def python(
        self, env_name: str | None = None, inherit: bool = True, default: Any = list
    ) -> list[str]:
        """Python getter"""
        out: list[str] = self._get_value(
            key="python",
            env_name=env_name,
            inherit=inherit,
            as_list=True,
            default=default,
        )
        return select_pythons(out, self.default_pythons, self.all_pythons)

    def _get_extras(
        self, key: str, env_name: str, default: Any, inherit: bool = True
    ) -> list[str]:
        val = self._get_value(
            key=key,
            env_name=env_name,
            inherit=inherit,
            default=default,
        )

        if isinstance(val, bool):
            return [env_name] if val else []

        if not isinstance(val, list):
            val = [val]
        return val  # type: ignore[no-any-return]

    def extras(self, env_name: str, inherit: bool = True) -> list[str]:
        """
        Extras getter

        * If value is `True` (default), then return [env_name]
        * If value is `False`, return []
        * else return list of extras
        """
        return self._get_extras(
            key="extras", env_name=env_name, default=list, inherit=inherit
        )

    def groups(self, env_name: str, inherit: bool = True) -> list[str]:
        """
        Groups getter.

        Same style as `self.extras`
        """
        return self._get_extras(
            key="groups", env_name=env_name, default=list, inherit=inherit
        )

    def extras_or_groups(self, env_name: str, inherit: bool = True) -> list[str]:
        """
        Extras_or_Groups getter.

        These will need to be resolved after the fact.
        """
        return self._get_extras(
            key="extras_or_groups", env_name=env_name, default=list, inherit=inherit
        )

    def output(self, env_name: str | None = None) -> str | None:
        """Output getter"""
        return self._get_value(key="output", env_name=env_name, inherit=False)  # type: ignore[no-any-return]

    def sort(
        self, env_name: str | None = None, inherit: bool = True, default: bool = True
    ) -> bool:
        """Sort getter"""
        return self._get_value(  # type: ignore[no-any-return]
            key="sort", env_name=env_name, inherit=inherit, default=default
        )

    def skip_package(self, env_name: str, default: bool = False) -> bool:
        """skip_package getter."""
        return self._get_value(key="skip_package", env_name=env_name, default=default)  # type: ignore[no-any-return]

    def name(self, env_name: str) -> bool:
        """Name option."""
        return self._get_value(key="name", env_name=env_name)  # type: ignore[no-any-return]

    def header(self, env_name: str) -> bool:
        """Header getter"""
        return self._get_value(key="header", env_name=env_name)  # type: ignore[no-any-return]

    def style(self, env_name: str | None = None, default: str = "yaml") -> str:
        """Style getter.  One of `yaml`, `requirements`"""
        out = self._get_value(
            key="style", env_name=env_name, default=default, as_list=True
        )
        for k in out:
            if k not in {"yaml", "requirements", "conda-requirements", "json"}:
                msg = f"unknown style {k}"
                raise ValueError(msg)
        return out  # type: ignore[no-any-return]

    def python_include(self, env_name: str | None = None) -> str | None:
        """Flag python_include"""
        return self._get_value(key="python_include", env_name=env_name)  # type: ignore[no-any-return]

    def python_version(self, env_name: str | None = None) -> str | None:
        """Flag python_version"""
        return self._get_value(key="python_version", env_name=env_name)  # type: ignore[no-any-return]

    def overwrite(self, env_name: str | None = None, default: str = "check") -> str:
        """Flag overwrite"""
        return self._get_value(key="overwrite", env_name=env_name, default=default)  # type: ignore[no-any-return]

    def verbose(
        self, env_name: str | None = None, default: int | None = None
    ) -> int | None:
        """Flag verbose"""
        return self._get_value(key="verbose", env_name=env_name, default=default)  # type: ignore[no-any-return]

    def template(self, env_name: str, default: str = "{env}") -> str:
        """Flag for template"""
        return self._get_value(key="template", env_name=env_name, default=default)  # type: ignore[no-any-return]

    def template_python(self, env_name: str, default: str = "py{py}-{env}") -> str:
        """Flag for template_python."""
        return self._get_value(  # type: ignore[no-any-return]
            key="template_python", env_name=env_name, default=default
        )

    def reqs_ext(self, env_name: str, default: str = ".txt") -> str:
        """Requirements extension"""
        return self._get_value(  # type: ignore[no-any-return]
            key="reqs_ext",
            env_name=env_name,
            default=default,
        )

    def yaml_ext(self, env_name: str, default: str = ".yaml") -> str:
        """Conda yaml extension"""
        return self._get_value(  # type: ignore[no-any-return]
            key="yaml_ext",
            env_name=env_name,
            default=default,
        )

    def deps(self, env_name: str, default: Any = None) -> list[str]:
        """Conda dependencies option."""
        return self._get_value(  # type: ignore[no-any-return]
            key="deps",
            env_name=env_name,
            default=default,
        )

    def reqs(self, env_name: str, default: Any = None) -> list[str]:
        """Pip dependencies option."""
        return self._get_value(  # type: ignore[no-any-return]
            key="reqs",
            env_name=env_name,
            default=default,
        )

    def user_config(self, env_name: str | None = None) -> str | None:  # noqa: ARG002
        """Flag user_config"""
        return self._get_value(key="user_config", default=None)  # type: ignore[no-any-return]

    def allow_empty(self, env_name: str | None = None, default: bool = False) -> bool:
        """Allow empty option."""
        return self._get_value(  # type: ignore[no-any-return]
            key="allow_empty", env_name=env_name, default=default
        )

    def remove_whitespace(
        self, env_name: str | None = None, default: bool = True
    ) -> bool:
        """Remove whitespace option."""
        return self._get_value(  # type: ignore[no-any-return]
            key="remove_whitespace",
            env_name=env_name,
            default=default,
        )

    def assign_user_config(self, user: Self) -> Self:
        """Assign user_config to self."""
        from copy import deepcopy

        data = deepcopy(self.data)

        # get user envs

        if "envs" not in data:
            data["envs"] = {}

        if "overrides" not in data:
            data["overrides"] = []

        for key in ["envs", "overrides"]:
            u = user.get_in(key)
            if u is not None:
                d = data[key]
                if isinstance(d, list):
                    if not isinstance(u, list):
                        msg = f"expected list, got {type(u)}"
                        raise TypeError(msg)
                    d.extend(u)  # pyright: ignore[reportUnknownMemberType]
                elif isinstance(d, dict):  # pragma: no cover
                    if not isinstance(u, dict):
                        msg = f"expected dict, got {type(u)}"
                        raise TypeError(msg)
                    d.update(**u)  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]

        return type(self)(data)

    def _get_output_and_templates(
        self, env_name: str, **defaults: Any
    ) -> list[str | None]:
        return [
            defaults.get(k, getattr(self, k)(env_name))
            for k in ["output", "template", "template_python"]
        ]

    def _iter_yaml(
        self, env_name: str, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        pythons = self.python(env_name)
        output, template, template_python = self._get_output_and_templates(
            env_name, **defaults
        )

        keys = [
            "extras",
            "groups",
            "extras_or_groups",
            "sort",
            "skip_package",
            "header",
            "overwrite",
            "verbose",
            "reqs",
            "deps",
            "name",
            "channels",
            "allow_empty",
            "remove_whitespace",
        ]

        data = {k: defaults.get(k, getattr(self, k)(env_name)) for k in keys}

        if not pythons:
            if output is None:
                output = filename_from_template(
                    template=template,
                    env_name=env_name,
                    ext=defaults.get("yaml_ext", self.yaml_ext(env_name)),
                )

            if python_include := self.python_include(env_name):
                data.update(python_include=python_include)

            if python_version := self.python_version(env_name):
                data.update(python_version=python_version)

            data.update(output=output)
            yield ("yaml", data)

        else:
            for python in pythons:
                output = filename_from_template(
                    template=template_python,
                    python=python,
                    env_name=env_name,
                    ext=defaults.get("yaml_ext", self.yaml_ext(env_name)),
                )
                yield ("yaml", dict(data, python=python, output=output))

    def _iter_reqs(
        self, env_name: str, remove_whitespace: bool | None = None, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        keys = [
            "extras",
            "groups",
            "extras_or_groups",
            "sort",
            "skip_package",
            "header",
            "overwrite",
            "verbose",
            "reqs",
            "allow_empty",
        ]

        output, template, _ = self._get_output_and_templates(env_name, **defaults)

        data = {k: defaults.get(k, getattr(self, k)(env_name)) for k in keys}

        # different default from yaml
        data["remove_whitespace"] = (
            remove_whitespace
            if remove_whitespace is not None
            else self.remove_whitespace(env_name, default=False)
        )

        output = self.output(env_name)
        if not output:  # pragma: no cover
            output = filename_from_template(
                template=template,
                env_name=env_name,
                ext=defaults.get("reqs_ext", self.reqs_ext(env_name)),
            )
        data.update(output=output)

        yield ("requirements", data)

    def iter_envs(
        self, envs: Sequence[str] | None = None, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        """Iterate over configs"""
        # filter defaults.  Only include values of not None:
        defaults = {k: v for k, v in defaults.items() if v is not None}

        if not envs:
            envs = list(self.envs.keys())

        for env in envs:
            for style in self.style(env):
                if style == "yaml":
                    yield from self._iter_yaml(env, **defaults)
                elif style == "requirements":
                    yield from self._iter_reqs(env, **defaults)
                else:  # pragma: no cover
                    msg = f"unknown style {style}"
                    raise ValueError(msg)

    @classmethod
    def from_toml_dict(
        cls, data_toml: dict[str, Any], user_config: dict[str, Any] | None = None
    ) -> Self:
        """Create from toml dictionaries."""
        data = get_in(["tool", "pyproject2conda"], data_toml, default={})

        if "default_envs" in data:
            default_envs = data.pop("default_envs")

            if "envs" not in data:
                data["envs"] = {}

            for env in default_envs:
                data["envs"][env] = {"extras_or_groups": True}

        c = cls(
            data,
            default_pythons=get_default_pythons(),
            all_pythons=get_all_pythons(data_toml),
        )

        # add in "default_envs"

        if user_config:  # pragma: no cover
            u = cls.from_toml_dict(user_config)
            c = c.assign_user_config(user=u)

        return c

    @classmethod
    def from_string(cls, s: str, user_config: str | None = None) -> Self:
        """Create from string representation of toml file."""
        from ._compat import tomllib

        c = cls.from_toml_dict(tomllib.loads(s))

        if user_config:
            u = cls.from_string(user_config)
            c = c.assign_user_config(user=u)
        return c

    @classmethod
    def from_file(
        cls, path: str | Path, user_config: str | Path | None = "infer"
    ) -> Self:
        """Create from toml file(s)."""
        from ._compat import tomllib

        with Path(path).open("rb") as f:
            data = tomllib.load(f)

        c = cls.from_toml_dict(data)

        if user_config == "infer" and (user_config := c.user_config()) is not None:
            # relative path
            user_config = Path(path).parent / Path(user_config)

        if user_config and Path(user_config).exists():
            u = cls.from_file(user_config)
            c = c.assign_user_config(u)

        return c



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from click.testing import CliRunner

if TYPE_CHECKING:
    from typing import Iterator


@pytest.fixture(scope="session")
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def example_path(tmp_path: Path) -> Iterator[Path]:
    old_cwd = Path.cwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(old_cwd)



"""Utilities for tools."""

from __future__ import annotations


def get_conda_environment_map(simplify: bool = True) -> dict[str, str]:
    """Construct mapping from environment env_name to path"""
    import subprocess
    from pathlib import Path

    result = subprocess.check_output(["conda", "env", "list"])

    env_map: dict[str, str] = {}
    for line in result.decode().split("\n"):
        if not line.startswith("#"):
            x = line.replace("*", "").split()
            if len(x) == 2:  # noqa: PLR2004
                env_map[x[0]] = x[1]
    if simplify:
        home = str(Path.home())
        env_map = {k: v.replace(home, "~") for k, v in env_map.items()}
    return env_map



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
import filecmp
import logging
import tempfile
from functools import partial
from pathlib import Path
from textwrap import dedent

import pytest

import pyproject2conda
from pyproject2conda import utils
from pyproject2conda.cli import app
from pyproject2conda.config import Config

ROOT = Path(__file__).resolve().parent / "data"


def do_run(runner, command, *opts, filename=None, must_exist=False):
    if filename is None:
        filename = str(ROOT / "test-pyproject.toml")
    filename = Path(filename)
    if must_exist and not filename.exists():
        msg = f"filename {filename} does not exist"
        raise ValueError(msg)

    return runner.invoke(app, [command, "-f", str(filename), *opts])


def test_template() -> None:
    assert utils.filename_from_template(None) is None

    expected = "py38-test.yaml"

    t = utils.filename_from_template(
        "py{py}-{env}", env_name="test", python="3.8", ext=".yaml"
    )

    assert t == expected

    t = utils.filename_from_template(
        "py{py}-{env}", env_name="test", python_version="3.8", ext=".yaml"
    )

    assert t == expected


@pytest.fixture
def simple_toml() -> str:
    return """\
    [project.optional-dependencies]
    test = [
        "pandas",
        "pytest", # p2c: -c conda-forge

    ]

    [dependency-groups]
    thing = ["a-thing"]

    [tool.pyproject2conda]
    channels = ['conda-forge']
    # these are the same as the default values of `p2c project`
    template_python = "py{py}-{env}"
    template = "hello-{env}"
    style = "yaml"
    # options
    python = "3.10"
    # Note that this is relative to the location of pyproject.toml
    user_config = "config/userconfig.toml"
    default_envs = ["test", "dev", "dist-pypi"]

    [tool.pyproject2conda.envs.base]
    extras = []
    style = "yaml"
    python = []

    [tool.pyproject2conda.envs.base2]
    style = "yaml"
    extras = []


    [tool.pyproject2conda.envs.base3]
    style = "yaml"
    python = "default"

    [tool.pyproject2conda.envs.base4]
    style = "yaml"
    python = ["3.9", "3.10", "3.11", "3.12", "3.13"]
    template_python = "py{py}-hello"

    [tool.pyproject2conda.envs.base5]
    style = "yaml"
    python = "all"
    template_python = "py{py}-hello"

    [tool.pyproject2conda.envs.base_lowest]
    style = "yaml"
    python = "lowest"
    template_python = "py{py}-hello"

    [tool.pyproject2conda.envs.base_highest]
    style = "yaml"
    python = "highest"
    template_python = "py{py}-hello"


    [tool.pyproject2conda.envs.extension_yaml]
    style = "yaml"
    yaml_ext = ".yml"

    [tool.pyproject2conda.envs.extension_txt]
    style = "requirements"
    reqs_ext = ".in"

    [tool.pyproject2conda.envs.both]
    groups = "thing"


    [[tool.pyproject2conda.overrides]]
    envs = ["both"]
    extras = ["test"]

    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"


    """


@pytest.fixture
def classifiers() -> str:
    return """
    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: Public Domain",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering",
    ]
    """


@pytest.fixture
def simple_config(simple_toml: str) -> Config:
    return Config.from_string(dedent(simple_toml))


def test_option_override_base(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["base"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "hello-base.yaml",
        },
    )


def test_option_override_base_reqs(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["base"], reqs=["-e ."]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": ["-e ."],
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "hello-base.yaml",
        },
    )


def test_option_override_base2(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["base2"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py310-base2.yaml",
            "python": "3.10",
        },
    )


def test_option_override_both(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["both"]))

    assert output[0] == (
        "yaml",
        {
            "extras": ["test"],
            "groups": ["thing"],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py310-both.yaml",
            "python": "3.10",
        },
    )


def test_option_override_base_template(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["base"], template="there-{env}"))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "there-base.yaml",
        },
    )


def test_option_override_base_allow_empty(simple_config: Config) -> None:
    output = list(
        simple_config.iter_envs(envs=["base"], allow_empty=True, template="there-{env}")
    )

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": True,
            "remove_whitespace": True,
            "output": "there-base.yaml",
        },
    )


def test_option_override_base_allow_empty_other(simple_config: Config) -> None:
    output = list(
        simple_config.iter_envs(
            envs=["base"],
            allow_empty=True,
            remove_whitespace=False,
            template="there-{env}",
        )
    )

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": True,
            "remove_whitespace": False,
            "output": "there-base.yaml",
        },
    )


def test_option_override_base3_default_python_error(
    example_path: Path,  # noqa: ARG001
    simple_config: Config,
) -> None:
    # using default python without a version
    with pytest.raises(ValueError, match="Must include `.python-version` .*"):
        list(simple_config.iter_envs(envs=["base3"]))


def test_option_override_base3_default_python(example_path, simple_toml: str) -> None:
    with (example_path / ".python-version").open("w") as f:
        f.write("3.10\n")

    from pyproject2conda.utils import get_default_pythons

    assert get_default_pythons() == ["3.10"]

    config = Config.from_string(dedent(simple_toml))
    output = list(config.iter_envs(envs=["base3"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py310-base3.yaml",
            "python": "3.10",
        },
    )


def test_option_override_all_pythons_error(simple_config: Config) -> None:
    with pytest.raises(ValueError, match="Must specify python versions .*"):
        list(simple_config.iter_envs(envs=["base5"]))


def test_option_override_all_pythons(simple_toml: str, classifiers: str) -> None:
    simple_config = Config.from_string(dedent(simple_toml + classifiers))

    a = list(simple_config.iter_envs(envs=["base4"]))
    b = list(simple_config.iter_envs(envs=["base5"]))

    assert len(a) == len(b) == 5

    assert a[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py39-hello.yaml",
            "python": "3.9",
        },
    )

    assert a == b


def test_option_override_lowest_highest(simple_toml: str, classifiers: str) -> None:
    simple_config = Config.from_string(dedent(simple_toml + classifiers))

    a = list(simple_config.iter_envs(envs=["base_lowest"]))
    b = list(simple_config.iter_envs(envs=["base_highest"]))

    assert a[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py39-hello.yaml",
            "python": "3.9",
        },
    )

    assert b[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py313-hello.yaml",
            "python": "3.13",
        },
    )


def test_option_override_extension_yaml(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["extension_yaml"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py310-extension_yaml.yml",
            "python": "3.10",
        },
    )


def test_option_override_extension_txt(simple_config: Config) -> None:
    output = list(simple_config.iter_envs(envs=["extension_txt"]))

    assert output[0] == (
        "requirements",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": [],
            "sort": True,
            "skip_package": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "allow_empty": False,
            "remove_whitespace": False,
            "output": "hello-extension_txt.in",
        },
    )


@pytest.mark.parametrize("fname", ["test-pyproject.toml", "test-pyproject-groups.toml"])
def test_dry(fname, runner) -> None:
    filename = ROOT / fname

    expected = """\
# --------------------
# Creating yaml py310-test-extras.yaml
channels:
  - conda-forge
dependencies:
  - python=3.10
  - conda-forge::pytest
  - pandas
# --------------------
# Creating yaml py311-test-extras.yaml
channels:
  - conda-forge
dependencies:
  - python=3.11
  - conda-forge::pytest
  - pandas
# --------------------
# Creating requirements test-extras.txt
pandas
pytest
    """

    result = do_run(
        runner, "project", "--dry", "--envs", "test-extras", filename=filename
    )

    assert result.output == dedent(expected)


def test_config_only_default() -> None:
    d0 = {
        "extras": [],
        "groups": [],
        "extras_or_groups": ["test"],
        "sort": True,
        "skip_package": False,
        "header": None,
        "overwrite": "check",
        "verbose": None,
        "name": None,
        "channels": None,
        "python": "3.8",
        "output": "py38-test.yaml",
        "deps": None,
        "reqs": None,
        "allow_empty": False,
        "remove_whitespace": True,
    }

    d1 = d0.copy()
    d1.update(extras=["test"], extras_or_groups=[])

    s0 = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]
    """
    s1 = """
    [tool.pyproject2conda.envs.test]
    python = ["3.8"]
    extras = true
    """

    for s, d in zip([s0, s1], (d0, d1)):
        c = Config.from_string(s)
        assert list(c.iter_envs()) == [("yaml", d)]


def test_config_errors() -> None:
    s = """
    [tool.pyproject2conda]
    python = ["3.8"]

    [tool.pyproject2conda.envs.test]
    extras = true
    """

    # raise error for bad env
    c = Config.from_string(s)
    with pytest.raises(ValueError):
        c.channels(env_name="hello")

    s1 = """
    [tool.pyproject2conda]
    python = ["3.8"]

    [tool.pyproject2conda.envs.test]
    style = "thing"
    """

    # raise error for bad env
    c = Config.from_string(s1)
    with pytest.raises(ValueError):
        c.style(env_name="test")

    with pytest.raises(ValueError):
        list(c.iter_envs())


def test_config_overrides() -> None:
    # test overrides env
    s = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]

    [[tool.pyproject2conda.overrides]]
    envs = ["test"]
    skip_package = true
    """

    c = Config.from_string(s)

    expected = (
        "yaml",
        {
            "extras": [],
            "groups": [],
            "extras_or_groups": ["test"],
            "sort": True,
            "skip_package": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "name": None,
            "channels": None,
            "python": "3.8",
            "output": "py38-test.yaml",
            "deps": None,
            "reqs": None,
            "allow_empty": False,
            "remove_whitespace": True,
        },
    )

    assert next(iter(c.iter_envs())) == expected


def test_conifg_overrides_no_envs() -> None:
    # test overrides env
    s = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]

    [[tool.pyproject2conda.overrides]]
    skip_package = true
    """

    c = Config.from_string(s)

    with pytest.raises(ValueError):
        c.overrides  # noqa: B018


def test_config_python_include_version() -> None:
    s = """
    [tool.pyproject2conda.envs.test-1]
    extras = ["test"]
    output = "py38-test.yaml"
    python_include = "3.8"
    python_version = "3.8"

    [tool.pyproject2conda.envs."py38-test"]
    extras = ["test"]
    python_include = "3.8"
    python_version = "3.8"
    """

    c = Config.from_string(s)

    expected = [
        (
            "yaml",
            {
                "extras": ["test"],
                "groups": [],
                "extras_or_groups": [],
                "sort": True,
                "skip_package": False,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python_include": "3.8",
                "python_version": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
        (
            "yaml",
            {
                "extras": ["test"],
                "groups": [],
                "extras_or_groups": [],
                "sort": True,
                "skip_package": False,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python_include": "3.8",
                "python_version": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
    ]

    assert list(c.iter_envs()) == expected

    # no output:


def test_config_user_config() -> None:
    # test overrides env
    s = """
    [tool.pyproject2conda.envs.test]
    python = "3.8"
    extras = "test"


    """

    s_user = """
    [tool.pyproject2conda.envs.user]
    extras = ["a", "b"]
    python = "3.9"

    [[tool.pyproject2conda.overrides]]
    envs = ["test"]
    skip_package = true
    """

    c = Config.from_string(s, s_user)

    expected = [
        (
            "yaml",
            {
                "extras": ["test"],
                "groups": [],
                "extras_or_groups": [],
                "sort": True,
                "skip_package": True,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
        (
            "yaml",
            {
                "extras": ["a", "b"],
                "groups": [],
                "extras_or_groups": [],
                "sort": True,
                "skip_package": False,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python": "3.9",
                "output": "py39-user.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
    ]

    assert list(c.iter_envs()) == expected

    # bad user
    s_user2 = """
    [[tool.pyproject2conda.envs]]
    extras = ["a", "b"]
    python = "3.9"

    [[tool.pyproject2conda.overrides]]
    envs = ["test"]
    skip_package = true
    """

    with pytest.raises(TypeError):
        c = Config.from_string(s, s_user2)

    s_user2 = """
    [tool.pyproject2conda.envs]
    extras = ["a", "b"]
    python = "3.9"

    [tool.pyproject2conda.overrides]
    envs = ["test"]
    skip_package = true
    """

    with pytest.raises(TypeError):
        c = Config.from_string(s, s_user2)

    # blank config, only user
    s2 = """
    [tool.pyproject2conda]
    """

    c = Config.from_string(s2, s_user)

    assert c.data == {
        "envs": {"user": {"extras": ["a", "b"], "python": "3.9"}},
        "overrides": [{"envs": ["test"], "skip_package": True}],
    }


def test_version(runner) -> None:
    result = runner.invoke(app, ["--version"])

    assert (
        result.stdout.strip()
        == f"pyproject2conda, version {pyproject2conda.__version__}"
    )


@pytest.mark.parametrize(
    ("fname", "opt"),
    [
        ("test-pyproject.toml", "-e"),
        ("test-pyproject-groups.toml", "-g"),
    ],
)
def test_multiple(fname, opt, runner, caplog) -> None:
    filename = ROOT / fname
    _do_run = partial(do_run, filename=filename)

    caplog.set_level(logging.INFO)

    t1 = tempfile.TemporaryDirectory()
    path1 = t1.name

    _do_run(
        runner,
        "project",
        "--template-python",
        f"{path1}/" + "py{py}-{env}",
        "--template",
        f"{path1}/" + "{env}",
    )

    assert "Creating" in caplog.text

    # running this again?
    _do_run(
        runner,
        "project",
        "-v",
        "--template-python",
        f"{path1}/" + "py{py}-{env}",
        "--template",
        f"{path1}/" + "{env}",
    )

    assert "Skipping requirements" in caplog.text

    # run again no verbose:
    _do_run(
        runner,
        "project",
        "--template-python",
        f"{path1}/" + "py{py}-{env}",
        "--template",
        f"{path1}/" + "{env}",
    )

    t2 = tempfile.TemporaryDirectory()
    path2 = t2.name

    _do_run(
        runner, "yaml", opt, "dev", "-p", "3.10", "-o", f"{path2}/py310-dev.yaml", "-v"
    )

    _do_run(
        runner,
        "yaml",
        opt,
        "dist-pypi",
        "--skip-package",
        "-p",
        "3.10",
        "-o",
        f"{path2}/py310-dist-pypi.yaml",
    )

    _do_run(runner, "yaml", opt, "test", "-p", "3.10", "-o", f"{path2}/py310-test.yaml")
    _do_run(runner, "yaml", opt, "test", "-p", "3.11", "-o", f"{path2}/py311-test.yaml")

    _do_run(
        runner,
        "yaml",
        opt,
        "test",
        "--skip-package",
        "-p",
        "3.10",
        "-o",
        f"{path2}/py310-test-extras.yaml",
    )
    _do_run(
        runner,
        "yaml",
        opt,
        "test",
        "--skip-package",
        "-p",
        "3.11",
        "-o",
        f"{path2}/py311-test-extras.yaml",
    )

    _do_run(
        runner, "r", opt, "test", "--skip-package", "-o", f"{path2}/test-extras.txt"
    )

    _do_run(
        runner,
        "yaml",
        opt,
        "dev",
        opt,
        "dist-pypi",
        "--name",
        "hello",
        "-p",
        "3.10",
        "-d",
        "extra-dep",
        "-r",
        "extra-req",
        "-o",
        f"{path2}/py310-user-dev.yaml",
    )

    _do_run(runner, "req", "-o", f"{path2}/base.txt")

    paths1 = Path(path1).glob("*")
    names1 = {x.name for x in paths1}

    expected = set(
        "base.txt py310-dev.yaml py310-dist-pypi.yaml py310-test-extras.yaml py310-test.yaml py310-user-dev.yaml py311-test-extras.yaml py311-test.yaml test-extras.txt".split()
    )

    assert names1 == expected

    paths2 = Path(path2).glob("*")
    names2 = {x.name for x in paths2}

    assert expected == names2

    for x in expected:
        assert filecmp.cmp(f"{path1}/{x}", f"{path2}/{x}")

    t1.cleanup()
    t2.cleanup()



"""Utilities to work with nox"""

from __future__ import annotations

import functools
import os
import shlex
from contextlib import contextmanager
from functools import cached_property, reduce
from pathlib import Path
from typing import TYPE_CHECKING, cast

# fmt: off
from nox.sessions import SessionRunner

DISALLOW_WHICH: list[str] = []

# * Override SessionRunner._create_venv ------------------------------------------------

_create_venv_super = SessionRunner._create_venv  # pyright: ignore[reportPrivateUsage]


def override_sessionrunner_create_venv(self: SessionRunner) -> None:
    """Override SessionRunner._create_venv"""
    if callable(self.func.venv_backend):
        # if passed a callable backend, always use just that
        logger.info("Using custom callable venv_backend")

        self.venv = self.func.venv_backend(self)  # pyright: ignore[reportAttributeAccessIssue]
        return None

    logger.info("Using nox venv_backend")
    return _create_venv_super(self)


SessionRunner._create_venv = override_sessionrunner_create_venv  # type: ignore[method-assign] # pyright: ignore[reportPrivateUsage]
# fmt: on

import locale
import operator

from nox.logger import logger
from nox.virtualenv import CondaEnv, VirtualEnv

if TYPE_CHECKING:
    import sys
    from typing import Any, Callable, Iterable, Iterator, Literal, Union

    from nox import Session

    PathLike = Union[str, Path]

    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self


def factory_conda_backend(
    backend: Literal["conda", "mamba", "micromamba"] = "conda",
    location: str | None = None,
) -> Callable[..., CondaEnv]:
    """Factory method for conda backend."""

    def passthrough_venv_backend(
        runner: SessionRunner,
    ) -> CondaEnv:
        # override allowed_globals
        CondaEnv.allowed_globals = ("conda", "mamba", "micromamba", "conda-lock")  # type: ignore[assignment]

        if not isinstance(runner.func.python, str):
            msg = "Python version is not a string"
            raise TypeError(msg)
        return CondaEnv(
            location=location or runner.envdir,
            interpreter=runner.func.python,
            reuse_existing=runner.func.reuse_venv
            or runner.global_config.reuse_existing_virtualenvs,
            venv_params=runner.func.venv_params,
            conda_cmd=backend,
        )

    return passthrough_venv_backend


def factory_virtualenv_backend(
    backend: Literal["virtualenv", "venv", "uv"] = "virtualenv",
    location: str | None = None,
) -> Callable[..., CondaEnv | VirtualEnv]:
    """Factory virtualenv backend."""

    def passthrough_venv_backend(
        runner: SessionRunner,
    ) -> VirtualEnv:
        venv = VirtualEnv(
            location=location or runner.envdir,
            interpreter=runner.func.python,  # type: ignore[arg-type]
            reuse_existing=runner.func.reuse_venv
            or runner.global_config.reuse_existing_virtualenvs,
            venv_params=runner.func.venv_params,
            venv_backend=backend,
        )
        venv.create()
        return venv

    return passthrough_venv_backend


# * Top level installation functions ---------------------------------------------------
@functools.cache
def cached_which(cmd: str) -> str | None:
    """
    Cached lookup of uv path.

    Returns path or None if not installed.
    """
    if cmd in DISALLOW_WHICH:
        return None

    from shutil import which

    return which(cmd)


def py_prefix(python_version: Any) -> str:
    """
    Get python prefix.

    `python="3.8` -> "py38"
    """
    if isinstance(python_version, str):
        return "py" + python_version.replace(".", "")
    msg = f"passed non-string value {python_version}"
    raise ValueError(msg)


def _verify_path(
    path: PathLike,
) -> Path:
    path = Path(path)
    if not path.is_file():
        msg = f"Path {path} is not a file"
        raise ValueError(msg)
    return path


def _verify_paths(
    paths: PathLike | Iterable[PathLike] | None,
) -> list[Path]:
    if paths is None:
        return []
    if isinstance(paths, (str, Path)):
        paths = [paths]
    return [_verify_path(p) for p in paths]


def infer_requirement_path_with_fallback(
    name: str | None,
    ext: str | None = None,
    python_version: str | None = None,
    lock: bool = False,
    check_exists: bool = True,
    lock_fallback: bool = False,
) -> tuple[bool, Path]:
    """Get the requirements file from options with fallback."""
    if lock_fallback:
        try:
            path = infer_requirement_path(
                name=name,
                ext=ext,
                python_version=python_version,
                lock=lock,
                check_exists=True,
            )
        except FileNotFoundError:
            logger.info("Falling back to non-locked")
            lock = False
            path = infer_requirement_path(
                name=name,
                ext=ext,
                python_version=python_version,
                lock=lock,
                check_exists=True,
            )

    else:
        path = infer_requirement_path(
            name=name,
            ext=ext,
            python_version=python_version,
            lock=lock,
            check_exists=check_exists,
        )
    return lock, path


def infer_requirement_path(
    name: str | None,
    ext: str | None = None,
    python_version: str | None = None,
    lock: bool = False,
    check_exists: bool = True,
) -> Path:
    """Get filename for a conda yaml or pip requirements file."""
    if name is None:
        msg = "must supply name"
        raise ValueError(msg)

    # adjust filename
    filename = name
    if ext is not None and not filename.endswith(ext):
        filename += ext
    if python_version is not None:
        prefix = py_prefix(python_version)
        if not filename.startswith(prefix):
            filename = f"{prefix}-{filename}"

    if lock:
        if filename.endswith(".yaml"):
            filename = filename.rstrip(".yaml") + "-conda-lock.yml"
        elif filename.endswith(".yml"):
            filename = filename.rstrip(".yml") + "-conda-lock.yml"
        elif filename.endswith(".txt"):
            pass
        else:
            msg = f"unknown file extension for {filename}"
            raise ValueError(msg)

        filename = f"./requirements/lock/{filename}"
    else:
        filename = f"./requirements/{filename}"

    path = Path(filename)
    if check_exists and not path.is_file():
        msg = f"{path} does not exist"
        raise FileNotFoundError(msg)

    return path


def _infer_requirement_paths(
    names: str | Iterable[str] | None,
    lock: bool = False,
    ext: str | None = None,
    python_version: str | None = None,
) -> list[Path]:
    if names is None:
        return []

    if isinstance(names, str):
        names = [names]
    return [
        infer_requirement_path(
            name,
            lock=lock,
            ext=ext,
            python_version=python_version,
        )
        for name in names
    ]


def is_conda_session(session: Session) -> bool:
    """Whether session is a conda session."""
    from nox.virtualenv import CondaEnv

    return isinstance(session.virtualenv, CondaEnv)


# * Main class ----------------------------------------------------------------
class Installer:
    """
    Class to handle installing package/dependencies

    Parameters
    ----------
    session : nox.Session
    update : bool, default=False
    lock : bool, default=False
    package: str or bool, optional
    pip_deps : str or list of str, optional
        pip dependencies
    requirements : str or list of str
        pip requirement file(s) (pip install -r requirements[0] ...)
        Can either be a full path or a envname (for example,
        "test" will get resolved to ./requirements/test.txt)
    constraints : str or list of str
        pip constraint file(s) (pip install -c ...)
    config_path :
        Where to save env config for future comparison.  Defaults to
        `session.virtualenv.location / "tmp" / "env.json"`.

    """

    def __init__(
        self,
        session: Session,
        *,
        update: bool = False,
        lock: bool = False,
        package: str | bool | None = None,
        pip_deps: str | Iterable[str] | None = None,
        requirements: PathLike | Iterable[PathLike] | None = None,
        constraints: PathLike | Iterable[PathLike] | None = None,
        config_path: PathLike | None = None,
        # conda specific things:
        conda_deps: str | Iterable[str] | None = None,
        conda_yaml: PathLike | None = None,
        create_venv: bool | None = False,
        channels: str | Iterable[str] | None = None,
    ) -> None:
        self.session = session

        self.update = update
        self.lock = lock

        if config_path is None:
            config_path = self.tmp_path / "env.json"
        else:
            config_path = Path(config_path)
        self.config_path = config_path

        if isinstance(package, bool):
            package = "-e . --no-deps" if package else None
        self.package = package

        self.pip_deps: list[str] = sorted(_remove_whitespace_list(pip_deps or []))

        self.requirements = sorted(_verify_paths(requirements))
        self.constraints = sorted(_verify_paths(constraints))

        # conda stuff
        self.conda_deps: list[str] = sorted(_remove_whitespace_list(conda_deps or []))
        self.channels: list[str] = sorted(_remove_whitespace_list(channels or []))

        if conda_yaml is not None:
            conda_yaml = _verify_path(conda_yaml)
        self.conda_yaml = conda_yaml

        if create_venv is None:
            create_venv = self.is_conda_session()
        self.create_venv = create_venv

        self._check_params()

    def _check_params(self) -> None:
        if not self.is_conda_session():
            if self.conda_deps or self.conda_yaml:
                msg = f"passing conda parameters to non conda session {self.conda_deps=} {self.conda_yaml=}"
                raise ValueError(
                    msg,
                )

            if self.lock and (
                not self.requirements or self.pip_deps or self.constraints
            ):
                msg = "Can only pass requirements for locked virtualenv"
                raise ValueError(msg)

        elif self.lock:
            if self.conda_yaml is None:
                msg = "Must pass `conda_yaml=conda-lock-file`"
                raise ValueError(msg)

            if (
                self.conda_deps
                or self.channels
                or self.pip_deps
                or self.requirements
                or self.constraints
            ):
                msg = "Can not pass conda_deps, channels, pip_deps, requirements, constraints if using conda-lock"
                raise ValueError(
                    msg,
                )

    @cached_property
    def config(self) -> dict[str, Any]:
        """Dictionary of relevant info for this session"""
        out: dict[str, Any] = {}
        out["lock"] = self.lock

        # special for package:
        if self.package:
            out["package"] = get_package_hash(self.package)

        for k in ["pip_deps", "conda_deps", "channels"]:
            if v := getattr(self, k):
                out[k] = v

        # file hashes
        for k in ["requirements", "constraints", "conda_yaml"]:
            if v := getattr(self, k):
                if isinstance(v, Path):
                    v = [v]
                out[k] = {str(path): _get_file_hash(path) for path in v}

        return out

    def save_config(self) -> Self:
        """Save config as json file to something like session/tmp/env.json"""
        import json

        # in case config path got clobbered
        self.config_path.parent.mkdir(parents=True, exist_ok=True)

        self.session.log(f"saving config to {self.config_path}")
        with self.config_path.open("w") as f:
            json.dump(self.config, f)
        return self

    @cached_property
    def previous_config(self) -> dict[str, Any]:
        """Previous config."""
        if not self.config_path.exists():
            return {}

        import json

        with self.config_path.open() as f:
            return json.load(f)  # type: ignore[no-any-return]

    @cached_property
    def tmp_path(self) -> Path:
        """
        Override session.create_tmp

        If override venv.location, then create tmp will
        not work correctly.  This will create a
        directory `venv.location / tmp`
        """
        return Path(self.session.virtualenv.location) / "tmp"

    def create_tmp_path(self) -> Path:
        """Create `self.tmp_path`"""
        tmp = self.tmp_path
        self.tmp_path.mkdir(parents=True, exist_ok=True)
        return tmp

    def log_session(self) -> Self:
        """Save log of session to `self.tmp_path / "env_info.txt"`."""
        logfile = Path(self.create_tmp_path()) / "env_info.txt"
        self.session.log(f"writing environment log to {logfile}")

        with logfile.open("w") as f:
            self.session.run("python", "--version", stdout=f)

            if self.is_conda_session():
                self.session.run("conda", "list", stdout=f, external=True)
            else:
                self.session.run("pip", "list", stdout=f)

        return self

    # Interface
    @property
    def python_version(self) -> str:
        """Python version for session."""
        return cast(str, self.session.python)

    def is_conda_session(self) -> bool:
        """Whether session is conda session."""
        return is_conda_session(self.session)

    @property
    def env(self) -> dict[str, str]:
        """Override environment variables"""
        if tmpdir := os.environ.get("TMPDIR"):
            return {"TMPDIR": tmpdir}
        return {}

    @property
    def conda_cmd(self) -> str:
        """Command for conda session (conda/mamba)."""
        venv = self.session.virtualenv
        if not isinstance(venv, CondaEnv):
            msg = "venv is not a CondaEnv"
            raise TypeError(msg)
        return venv.conda_cmd

    def is_micromamba(self) -> bool:
        """Whether conda session uses micromamba."""
        return self.is_conda_session() and self.conda_cmd == "micromamba"

    @cached_property
    def python_full_path(self) -> str:
        """Full path to session python executable."""
        path = self.session.run_always(
            "python",
            "-c",
            "import sys; print(sys.executable)",
            silent=True,
        )
        if not isinstance(path, str):
            msg = "accessing python_full_path with value None"
            raise TypeError(msg)
        return path.strip()

    @property
    def _session_runner(self) -> SessionRunner:
        return self.session._runner  # pyright: ignore[reportPrivateUsage]

    @cached_property
    def skip_install(self) -> bool:
        """Whether to skip install."""
        return self._session_runner.global_config.no_install and getattr(
            self._session_runner.venv,
            "_reused",
            False,
        )

    @cached_property
    def package_changed(self) -> bool:
        """Whether the package has changed"""
        if changed := self.config.get("package") != self.previous_config.get("package"):
            self.session.log("package changed")
        return changed

    @cached_property
    def changed(self) -> bool:
        """Check for changes (excluding package)"""
        a, b = (
            {k: v for k, v in config.items() if k != "package"}
            for config in (self.config, self.previous_config)
        )

        out = a != b

        msg = "changed" if out else "unchanged"
        self.session.log(f"session {self.session.name} {msg}")

        return out

    # Smart runners
    def run_commands(
        self,
        commands: Iterable[str | Iterable[str]] | None,
        external: bool = True,
        **kwargs: Any,
    ) -> Self:
        """Run commands in session."""
        if commands:
            kwargs.update(external=external)
            for opt in combine_list_list_str(commands):
                self.session.run(*opt, **kwargs)
        return self

    def set_ipykernel_display_name(
        self,
        name: str,
        display_name: str | None = None,
        user: bool = True,
        update: bool = False,
    ) -> Self:
        """Set ipykernel display name."""
        if self.changed or update or self.update:
            if display_name is None:
                display_name = f"Python [venv: {name}]"
            self.session.run_always(
                "python",
                "-m",
                "ipykernel",
                "install",
                "--user" if user else "--sys-prefix",
                "--name",
                name,
                "--display-name",
                display_name,
                success_codes=[0, 1],
            )

        return self

    def install_all(
        self,
        update_package: bool = False,
        log_session: bool = False,
        save_config: bool = True,
    ) -> Self:
        """Install package/dependencies."""
        if self.create_venv:
            if not self.is_conda_session():
                msg = "Only CondaEnv should be used with create_venv"
                raise TypeError(msg)
            self.create_conda_env()

        out = (
            (self.conda_install_deps() if self.is_conda_session() else self)
            .pip_install_deps()
            .pip_install_package(update=update_package)
        )

        if log_session:
            out = out.log_session()

        if save_config:
            out = out.save_config()

        return out

    def uv_install(self, *args: str, **kwargs: Any) -> None:
        """Run uv pip install if available"""
        if uv_path := cached_which("uv"):
            self.session.run_always(
                uv_path,
                "pip",
                "install",
                f"--python={self.python_full_path}",
                *args,
                **kwargs,
                external=True,
            )
        else:
            self.session.install(*args, **kwargs)

    def pip_install_package(
        self,
        *args: Any,
        update: bool = False,
        opts: str | list[str] | None = None,
        **kwargs: Any,
    ) -> Self:
        """Install the package"""
        if self.package is None or self.skip_install:
            pass

        elif self.changed or self.package_changed or (update := self.update or update):
            command = shlex.split(self.package, posix=True)

            if update:
                command.append("--upgrade")

            if opts:
                command.extend(combine_list_str(opts))

            self.uv_install(*command, *args, **kwargs)

        return self

    def pip_install_deps(
        self,
        *args: Any,
        update: bool = False,
        opts: str | Iterable[str] | None = None,
        **kwargs: Any,
    ) -> Self:
        """Install pip dependencies with pip or pip-sync."""
        if self.skip_install:
            pass

        elif self.changed or (update := update or self.update):
            # # Playing with using pip-sync.
            # # Not sure its worth it?
            if self.lock:
                # Using pip-compile-{python_version} session.
                if not isinstance(self.session.python, str):
                    raise TypeError

                if uv_path := cached_which("uv"):
                    self.session.run_always(
                        uv_path,
                        "pip",
                        "sync",
                        f"--python={self.python_full_path}",
                        *map(str, self.requirements),
                        external=True,
                    )
                else:
                    self.session.run_always(
                        "nox",
                        "-s",
                        f"pip-compile-{self.session.python}",
                        "--",
                        "++pip-compile-run-internal",
                        "pip-sync",
                        "--python-executable",
                        self.python_full_path,
                        *map(str, self.requirements),
                        silent=True,
                        external=True,
                    )

            else:
                install_args: list[str] = (
                    prepend_flag("-r", *map(str, self.requirements))
                    + prepend_flag("-c", *map(str, self.constraints))
                    + self.pip_deps
                )

                if install_args:
                    if update:
                        install_args = ["--upgrade", *install_args]

                    if opts:
                        install_args.extend(combine_list_str(opts))
                    self.uv_install(*install_args, *args, **kwargs)

        return self

    def create_conda_env(self, update: bool = False) -> Self:
        """Create conda environment."""
        venv = self.session.virtualenv
        if not isinstance(venv, CondaEnv):
            msg = "Session must be a conda session."
            raise TypeError(msg)

        if venv._clean_location():  # pyright: ignore[reportPrivateUsage]
            # Also clean out session tmp directory
            # shutil.rmtree(self.session.create_tmp())  # noqa: ERA001
            cmd = "create"
        elif self.changed or update or self.update:
            cmd = "update"
        else:
            venv._reused = True  # pyright: ignore[reportPrivateUsage]
            cmd = "reuse"

        cmds = [cmd]
        if cmd == "update":
            cmds.append("--prune")

        # create environment
        self.session.log(
            f"{cmd.capitalize()} conda environment in {venv.location_name}",
        )

        if cmd != "reuse":
            extra_params: list[str] = self._session_runner.func.venv_params or []

            if self.lock and not self.is_micromamba():
                # use conda-lock
                if self.conda_cmd == "conda":
                    extra_params.append("--no-mamba")
                else:
                    extra_params.append("--mamba")

                extra_params.extend(["--prefix", venv.location])

                cmds = ["conda-lock", "install", *extra_params, str(self.conda_yaml)]
            else:
                cmds = (
                    [self.conda_cmd]
                    + ([] if self.is_micromamba() else ["env"])
                    + cmds
                    + (["--yes"] if self.is_micromamba() else [])
                    + [
                        "--prefix",
                        venv.location,
                        "--file",
                        str(self.conda_yaml),
                    ]
                    + extra_params
                )

            self.session.run_always(*cmds, silent=True, env=self.env)

        return self

    def conda_install_deps(
        self,
        update: bool = False,
        opts: str | Iterable[str] | None = None,
        channels: str | Iterable[str] | None = None,
        prune: bool = False,
        **kwargs: Any,
    ) -> Self:
        """Install conda dependencies (apart from environment.yaml)."""
        if (not self.conda_deps) or self.skip_install:
            pass

        elif self.changed or (update := (update or self.update)):
            channels = channels or self.channels
            if channels:
                kwargs.update(channel=channels)

            deps = list(self.conda_deps)

            if update and not self.is_micromamba():
                deps.insert(0, "--update-all")

            if prune and not self.is_micromamba():
                deps.insert(0, "--prune")

            if (update or prune) and self.is_micromamba():
                self.session.warn(
                    "Trying to run update with micromamba.  You should rebuild the session instead.",
                )

            if opts:
                deps.extend(combine_list_str(opts))

            self.session.conda_install(*deps, **kwargs, env=self.env)

        return self

    @classmethod
    def from_envname_pip(
        cls,
        session: Session,
        envname: str | Iterable[str] | None = None,
        lock: bool = False,
        requirements: PathLike | Iterable[PathLike] | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create object for virtualenv from envname."""
        if lock:
            if not isinstance(session.python, str):
                msg = "session.python must be a string"
                raise TypeError(msg)
            requirements = _verify_paths(requirements) + _infer_requirement_paths(
                envname,
                ext=".txt",
                lock=lock,
                python_version=session.python,
            )

        elif envname is not None:
            requirements = _verify_paths(requirements) + _infer_requirement_paths(
                envname,
                ext=".txt",
            )

        return cls(
            session=session,
            lock=lock,
            requirements=requirements,
            **kwargs,
        )

    @classmethod
    def from_envname_conda(
        cls,
        session: Session,
        envname: str | None = None,
        conda_yaml: PathLike | None = None,
        lock: bool = False,
        lock_fallback: bool = True,
        conda_deps: str | Iterable[str] | None = None,
        pip_deps: str | Iterable[str] | None = None,
        channels: str | Iterable[str] | None = None,
        package: str | None = None,
        create_venv: bool | None = None,
        **kwargs: Any,
    ) -> Self:
        """
        Create object for conda environment from envname.

        Parameters
        ----------
        envname :
            Base name for file.  For example, passing
            envname = "dev" will convert to
            `requirements/py{py}-dev.yaml` for `filename`

        """
        if envname is not None and conda_yaml is None:
            if not isinstance(session.python, str):
                msg = "session.python must be a string"
                raise TypeError(msg)
            lock, conda_yaml = infer_requirement_path_with_fallback(
                envname,
                ext=".yaml",
                python_version=session.python,
                lock=lock,
                lock_fallback=lock_fallback,
            )
        elif envname is None and conda_yaml is not None:
            conda_yaml = _verify_path(conda_yaml)

        else:
            msg = "Pass one of envname or conda_yaml"
            raise ValueError(msg)

        return cls(
            session=session,
            lock=lock,
            conda_yaml=conda_yaml,
            conda_deps=conda_deps,
            pip_deps=pip_deps,
            channels=channels,
            package=package,
            create_venv=create_venv,
            **kwargs,
        )

    @classmethod
    def from_envname(
        cls,
        session: Session,
        envname: str | Iterable[str] | None = None,
        lock: bool = False,
        lock_fallback: bool | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create object from envname."""
        if lock_fallback is None:
            lock_fallback = is_conda_session(session)

        if is_conda_session(session):
            if not isinstance(envname, str) and envname is not None:
                msg = "envname must be a string or None"
                raise TypeError(msg)
            return cls.from_envname_conda(
                session=session,
                envname=envname,
                lock=lock,
                lock_fallback=lock_fallback,
                **kwargs,
            )

        return cls.from_envname_pip(
            session=session,
            envname=envname,
            lock=lock,
            **kwargs,
        )


# * Utilities --------------------------------------------------------------------------
def _remove_whitespace(s: str) -> str:
    import re

    return re.sub(r"\s+", "", s)


def _remove_whitespace_list(s: str | Iterable[str]) -> list[str]:
    if isinstance(s, str):
        s = [s]
    return [_remove_whitespace(x) for x in s]


def combine_list_str(opts: str | Iterable[str]) -> list[str]:
    """Cleanup str/list[str] to list[str]"""
    if not opts:
        return []

    if isinstance(opts, str):
        opts = [opts]
    return shlex.split(" ".join(opts))


def combine_list_list_str(opts: Iterable[str | Iterable[str]]) -> Iterable[list[str]]:
    """Cleanup Iterable[str/list[str]] to Iterable[list[str]]."""
    return (combine_list_str(opt) for opt in opts)


def sort_like(values: Iterable[Any], like: Iterable[Any]) -> list[Any]:
    """Sort `values` in order of `like`."""
    # only unique
    sorter = {k: i for i, k in enumerate(like)}
    return sorted(set(values), key=lambda k: sorter[k])


def update_target(
    target: str | Path,
    *deps: str | Path,
    allow_missing: bool = False,
) -> bool:
    """Check if target is older than deps:"""
    target = Path(target)
    if not target.exists():
        return True

    deps_filtered: list[Path] = []
    for d in map(Path, deps):
        if d.exists():
            deps_filtered.append(d)
        elif not allow_missing:
            msg = f"dependency {d} does not exist"
            raise ValueError(msg)

    target_time = target.stat().st_mtime
    return any(target_time < dep.stat().st_mtime for dep in deps_filtered)


def prepend_flag(flag: str, *args: str | Iterable[str]) -> list[str]:
    """
    Add in a flag before each arg.

    >>> prepent_flag("-k", "a", "b")
    ["-k", "a", "-k", "b"]
    """
    args_: list[str] = []
    for x in args:
        if isinstance(x, str):
            args_.append(x)
        else:
            args_.extend(x)

    return reduce(operator.iadd, [[flag, _] for _ in args_], [])


def _get_file_hash(path: str | Path, buff_size: int = 65536) -> str:
    import hashlib

    md5 = hashlib.md5()
    with Path(path).open("rb") as f:
        while True:
            data = f.read(buff_size)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def _get_zipfile_hash(path: str | Path) -> str:
    import hashlib
    from zipfile import ZipFile

    md5 = hashlib.md5()

    with ZipFile(path, "r") as zipfile:
        for item in sorted(zipfile.infolist(), key=lambda x: x.filename):
            md5.update(item.filename.encode())
            md5.update(zipfile.read(item))

    return md5.hexdigest()


def get_package_hash(package: str) -> list[str]:
    """Get hash for wheel of package."""
    import re

    out: list[str] = []

    for x in shlex.split(package):
        out.append(x)

        # remove possible extras
        x_clean = re.sub(r"\[.*?\]$", "", x)
        # remove possible thing@path -> path
        x_clean = re.sub(r"^.*?\@", "", x_clean)

        if Path(x_clean).is_file():
            if x_clean.endswith(".whl"):
                out.append(_get_zipfile_hash(x_clean))
            else:
                out.append(_get_file_hash(x_clean))

    return out


@contextmanager
def check_for_change_manager(
    *deps: str | Path,
    hash_path: str | Path | None = None,
    target_path: str | Path | None = None,
    force_write: bool = False,
) -> Iterator[bool]:
    """
    Context manager to look for changes in dependencies.

    Yields
    ------
    changed: bool

    If exit normally, write hashes to hash_path file

    """
    try:
        changed, hashes, hash_path = check_hash_path_for_change(
            *deps,
            target_path=target_path,
            hash_path=hash_path,
        )

        yield changed

    except Exception:  # noqa: TRY203
        raise

    else:
        if force_write or changed:
            logger.info(f"Writing {hash_path}")

            # make sure the parent directory exists:
            hash_path.parent.mkdir(parents=True, exist_ok=True)
            write_hashes(hash_path=hash_path, hashes=hashes)


def check_hash_path_for_change(
    *deps: str | Path,
    target_path: str | Path | None = None,
    hash_path: str | Path | None = None,
) -> tuple[bool, dict[str, str], Path]:
    """
    Checks a json file `hash_path` for hashes of `other_paths`.

    if specify target_path and no hash_path, set `hash_path=target_path.parent / (target_path.name + ".hash.json")`.
    if specify hash_path and no target, set

    Parameters
    ----------
    *deps :
        files on which target_path/hash_path depends.
    hash_path :
        Path of file containing hashes of `deps`.
    target_path :
        Target file (i.e., the final file to be created).
        Defaults to hash_path.


    Returns
    -------
    changed : bool
    hashes : dict[str, str]
    hash_path : Path

    """
    import json

    msg = "Must specify target_path or hash_path"

    if target_path is None:
        if hash_path is None:
            raise ValueError(msg)
        target_path = hash_path = Path(hash_path)
    else:
        target_path = Path(target_path)
        if hash_path is None:
            hash_path = target_path.parent / (target_path.name + ".hash.json")
        else:
            hash_path = Path(hash_path)

    hashes: dict[str, str] = {
        os.path.relpath(k, hash_path.parent): _get_file_hash(k) for k in deps
    }

    if not target_path.is_file():
        changed = True

    elif hash_path.is_file():
        with hash_path.open() as f:
            previous_hashes: dict[str, Any] = json.load(f)

        changed = False
        for k, h in hashes.items():
            previous = previous_hashes.get(k)
            if previous is None or previous != h:
                changed = True
                hashes = {**previous_hashes, **hashes}
                break

    else:
        changed = True

    return changed, hashes, hash_path


def write_hashes(hash_path: str | Path, hashes: dict[str, Any]) -> None:
    """Write hashes to json file."""
    import json

    with Path(hash_path).open("w", encoding=locale.getpreferredencoding(False)) as f:
        json.dump(hashes, f, indent=2)
        f.write("\n")


def open_webpage(path: str | Path | None = None, url: str | None = None) -> None:
    """
    Open webpage from path or url.

    Useful if want to view webpage with javascript, etc., as well as static html.
    """
    import webbrowser
    from urllib.request import pathname2url

    if path:
        url = "file://" + pathname2url(str(Path(path).absolute()))
    if url:
        webbrowser.open(url)


def session_run_commands(
    session: Session,
    commands: list[list[str]] | None,
    external: bool = True,
    **kws: Any,
) -> None:
    """Run commands command."""
    if commands:
        kws.update(external=external)
        for opt in combine_list_list_str(commands):
            session.run(*opt, **kws)


# * User config ------------------------------------------------------------------------
def add_uv_pythons_to_path() -> None:
    """Add paths to uv python"""
    import os
    import subprocess
    from pathlib import Path

    try:
        path_to_uv_pythons = (
            subprocess.check_output(
                ["uv", "python", "dir"],
                env={"NO_COLOR": "1", **os.environ},
            )
            .decode()
            .strip()
        )
        paths_new = ":".join(map(str, Path(path_to_uv_pythons).glob("*/bin")))
        paths_old = os.environ["PATH"]
        os.environ["PATH"] = f"{paths_new}:{paths_old}"
    except subprocess.CalledProcessError:
        pass


def load_nox_config(path: str | Path = "./config/userconfig.toml") -> dict[str, Any]:
    """
    Load user toml config file.

    File should look something like:

    [nox.python]
    paths = ["~/.conda/envs/python-3.*/bin"]

    # Extras for environments
    # for example, could have
    # dev = ["dev", "nox", "tools"]
    [nox.extras]
    dev = ["dev", "nox"]
    """
    from .projectconfig import ProjectConfig

    return ProjectConfig.from_path_and_environ(path).to_nox_config()



"""
Override dependencies (:mod:`~pyproject2conda.overrides`)
=========================================================
"""

from __future__ import annotations

from typing import TypedDict


class OverrideDict(TypedDict, total=False):
    """Dict for storing override options."""

    pip: bool
    skip: bool
    channel: str | None
    packages: str | list[str]


class OverrideDeps:
    """Class to work with overrides from comment or table"""

    def __init__(
        self,
        pip: bool = False,
        skip: bool = False,
        packages: str | list[str] | None = None,
        channel: str | None = None,
    ) -> None:
        if channel is not None and channel.strip() in {"pip", "pypi"}:
            channel = None
            pip = True

        self.pip = pip
        self.skip = skip
        self.channel = channel

        if packages is None:
            packages = []
        elif isinstance(packages, str):
            packages = [packages]
        self.packages = packages

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self.__dict__)



"""
Requirements parsing (:mod:`~pyproject2conda.requirements`)
===========================================================
"""

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Iterable,
        Sequence,
    )

    from ._typing import (
        MISSING_TYPE,
    )
    from ._typing_compat import Self

from copy import copy

from packaging.markers import Marker
from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import SpecifierSet

from pyproject2conda.utils import (
    MISSING,
    get_in,
    list_to_str,
    unique_list,
)
from pyproject2conda.utils import (
    remove_whitespace_list as _remove_whitespace_list,
)

from .overrides import OverrideDeps


# * Utilities --------------------------------------------------------------------------
def _check_allow_empty(allow_empty: bool) -> str:
    msg = "No dependencies for this environment\n"
    if allow_empty:
        return msg
    raise ValueError(msg)


def _clean_pip_reqs(reqs: list[str]) -> list[str]:
    out: list[str] = []
    for req in reqs:
        try:
            r = str(Requirement(req))
        except InvalidRequirement:
            # trust that user knows what they're doing
            r = req
        out.append(r)

    return out


def _clean_conda_requirement(
    requirement: Requirement,
    python_version: str | None = None,
    channel: str | None = None,
) -> Requirement | None:
    if (
        python_version
        and requirement.marker
        and (not requirement.marker.evaluate({"python_version": python_version}))
    ):
        return None

    requirement = _update_requirement(requirement, marker=None, extras=None)
    if channel:
        requirement.name = f"{channel}::{requirement.name}"
    return requirement


def _clean_conda_strings(
    deps: list[str], python_version: str | None = None
) -> list[str]:
    out: list[str] = []
    for dep in deps:
        # if have a channel, take it out
        if "::" in dep:
            channel, d = dep.split("::")
        else:
            channel, d = None, dep

        r = _clean_conda_requirement(
            Requirement(d), python_version=python_version, channel=channel
        )
        if r is not None:
            out.append(str(r))
    return out


def _update_requirement(  # noqa: C901, PLR0912
    requirement: str | Requirement,
    name: str | MISSING_TYPE = MISSING,
    url: str | None | MISSING_TYPE = MISSING,
    extras: str | Iterable[str] | None | MISSING_TYPE = MISSING,
    specifier: str | SpecifierSet | None | MISSING_TYPE = MISSING,
    marker: str | Marker | None | MISSING_TYPE = MISSING,
) -> Requirement:  # pragma: no cover
    if isinstance(requirement, str):
        requirement = Requirement(requirement)
    else:
        requirement = copy(requirement)

    if name is not MISSING:
        requirement.name = name

    if url is not MISSING:
        requirement.url = url

    if extras is not MISSING:
        if extras is None:
            extras = set()
        elif isinstance(extras, str):
            extras = {extras}
        else:
            extras = set(extras)
        requirement.extras = extras

    if specifier is not MISSING:
        if specifier is None:
            specifier = SpecifierSet()
        elif isinstance(specifier, str):
            specifier = SpecifierSet(specifier)
        requirement.specifier = specifier

    if marker is not MISSING:
        if isinstance(marker, str):
            marker = Marker(marker)

        requirement.marker = marker

    return requirement


# ** Dependencices
def resolve_extras(
    extras: str | Iterable[str],
    package_name: str,
    unresolved: dict[str, list[Requirement]],
) -> list[Requirement]:
    """Resolve extras"""
    if isinstance(extras, str):
        extras = [extras]

    out: list[Requirement] = []
    for extra in extras:
        for requirement in unresolved[extra]:
            if requirement.name == package_name:
                out.extend(
                    resolve_extras(
                        extras=requirement.extras,
                        package_name=package_name,
                        unresolved=unresolved,
                    )
                )
            else:
                out.append(requirement)
    return out


# ** output ----------------------------------------------------------------------------
def _conda_yaml(
    name: str | None = None,
    channels: str | Iterable[str] | None = None,
    conda_deps: str | Iterable[str] | None = None,
    pip_deps: str | Iterable[str] | None = None,
    add_file_eol: bool = True,
) -> str:
    def _as_list(x: str | Iterable[str]) -> Iterable[str]:
        if isinstance(x, str):
            return [x]
        return x

    if not conda_deps:
        msg = "Must have at least one conda dependency (i.e., pip)"
        raise ValueError(msg)

    out: list[str] = []
    if name is not None:
        out.append(f"name: {name}")

    if channels:
        out.append("channels:")
        out.extend(f"  - {channel}" for channel in _as_list(channels))

    out.append("dependencies:")
    out.extend(f"  - {dep}" for dep in _as_list(conda_deps))

    if pip_deps:
        if "pip" not in conda_deps:  # pragma: no cover
            msg = "Must have pip in conda_deps"
            raise ValueError(msg)
        out.append("  - pip:")
        out.extend(f"      - {dep}" for dep in _as_list(pip_deps))

    s = "\n".join(out)

    if add_file_eol:  # pragma: no cover
        s += "\n"

    return s


def _create_header(cmd: str | None = None) -> str:
    from textwrap import dedent

    if cmd:
        header = dedent(
            f"""
        This file is autogenerated by pyproject2conda
        with the following command:

            $ {cmd}

        You should not manually edit this file.
        Instead edit the corresponding pyproject.toml file.
        """
        )
    else:
        header = dedent(
            """
        This file is autogenerated by pyproject2conda.
        You should not manually edit this file.
        Instead edit the corresponding pyproject.toml file.
        """
        )

    # prepend '# '
    lines: list[str] = []
    for line in header.split("\n"):
        if len(line.strip()) == 0:
            lines.append("#")
        else:
            lines.append("# " + line)
    return "\n".join(lines)


def _add_header(string: str, header_cmd: str | None) -> str:
    if header_cmd is not None:
        return _create_header(header_cmd) + "\n" + string
    return string


def _optional_write(
    string: str,
    output: str | Path | None,
    mode: str = "w",  # force: bool = False,
) -> None:
    if output is None:
        return

    path = Path(output)

    with path.open(mode) as f:
        f.write(string)


# * Main class
class ParseDepends:
    """
    Parse pyproject.toml file for dependencies

    Parameters
    ----------
    data : dict
    """

    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def get_in(
        self, *keys: str, default: Any = None, factory: Callable[[], Any] | None = None
    ) -> Any:
        """Generic getter."""
        return get_in(
            keys=keys, nested_dict=self.data, default=default, factory=factory
        )

    @cached_property
    def package_name(self) -> str:
        """Clean name of package."""
        if (out := self.get_in("project", "name")) is None:
            msg = "Must specify `project.name`"
            raise ValueError(msg)
        return cast(str, out)

    @cached_property
    def dependencies(self) -> list[str]:
        """project.dependencies"""
        return cast(
            "list[str]",
            self.get_in(
                "project",
                "dependencies",
                factory=list,
            ),
        )

    @cached_property
    def build_system_requires(self) -> list[str]:
        """build-system.requires"""
        return cast(
            "list[str]",
            self.get_in(
                "build-system",
                "requires",
                factory=list,
            ),
        )

    @cached_property
    def optional_dependencies(self) -> dict[str, Any]:
        """project.optional-dependencies"""
        return cast(
            "dict[str, Any]",
            self.get_in(
                "project",
                "optional-dependencies",
                factory=dict,
            ),
        )

    @cached_property
    def dependency_groups(self) -> dict[str, Any]:
        """dependency-groups"""
        return cast(
            "dict[str, Any]",
            self.get_in(
                "dependency-groups",
                factory=dict,
            ),
        )

    @cached_property
    def override_table(self) -> dict[str, OverrideDeps]:
        """
        tool.pyproject2conda.dependencies

        Mapping from requirement name to OverrideDeps instance

        """
        out = self.get_in("tool", "pyproject2conda", "dependencies", factory=dict)
        if out:
            out = {k: OverrideDeps(**v) for k, v in out.items()}
        return cast("dict[str, OverrideDeps]", out)

    @cached_property
    def channels(self) -> list[str]:
        """tool.pyproject2conda.channels"""
        channels = self.get_in("tool", "pyproject2conda", "channels", factory=list)

        if isinstance(channels, str):
            channels = [channels]
        return channels  # type: ignore[no-any-return]

    @property
    def extras(self) -> list[str]:
        """Available extras"""
        return [*self.optional_dependencies, "build-system.requires"]

    @property
    def groups(self) -> list[str]:
        """Available groups"""
        return [*self.dependency_groups, "build-system.requires"]

    @cached_property
    def requirements_base(self) -> list[Requirement]:
        """Base requirements"""
        return [Requirement(x) for x in self.dependencies]

    @cached_property
    def requirements_extras(self) -> dict[str, list[Requirement]]:
        """Extras requirements"""
        unresolved: dict[str, list[Requirement]] = {
            k: [Requirement(x) for x in v]
            for k, v in self.optional_dependencies.items()
        }

        resolved = {
            extra: resolve_extras(
                extra, package_name=self.package_name, unresolved=unresolved
            )
            for extra in unresolved
        }

        # add in build-system.requires
        resolved["build-system.requires"] = [
            Requirement(x) for x in self.build_system_requires
        ]

        return resolved

    @cached_property
    def requirements_groups(self) -> dict[str, list[Requirement]]:
        """Groups requirements"""
        from .vendored.dependency_groups._implementation import resolve

        resolved: dict[str, list[Requirement]] = {
            group: [Requirement(x) for x in resolve(self.dependency_groups, group)]
            for group in self.dependency_groups
        }

        # add in build-system.requires
        resolved["build-system.requires"] = [
            Requirement(x) for x in self.build_system_requires
        ]

        return resolved

    @staticmethod
    def _check_prop(vals: str | Iterable[str] | None, keys: list[str]) -> list[str]:
        if vals is None:
            return []

        if isinstance(vals, str):
            vals = [vals]

        out: list[str] = []
        for val in vals:
            if val not in keys:
                msg = f"{vals} not in {keys}"
                raise ValueError(msg)
            out.append(val)

        return out

    def _resolve_extras_and_groups(
        self,
        extras: str | Iterable[str] | None,
        groups: str | Iterable[str] | None,
        extras_or_groups: str | Iterable[str] | None,
    ) -> tuple[list[str], list[str]]:
        """Update extras and groups from extras_or_groups."""
        extras = self._check_prop(extras, self.extras)
        groups = self._check_prop(groups, self.groups)

        if extras_or_groups:
            if isinstance(extras_or_groups, str):
                extras_or_groups = [extras_or_groups]

            for extra_or_group in extras_or_groups:
                if extra_or_group in self.extras:
                    extras.append(extra_or_group)
                elif extra_or_group in self.groups:
                    groups.append(extra_or_group)
                else:  # pragma: no cover
                    msg = f"extra-or-group {extra_or_group} not in extras or groups"
                    raise ValueError(msg)

        return extras, groups

    @staticmethod
    def _cleanup(
        values: list[str],
        remove_whitespace: bool = True,
        unique: bool = True,
        sort: bool = True,
    ) -> list[str]:
        if remove_whitespace:
            values = _remove_whitespace_list(values)

        if unique:  # pragma: no cover
            values = unique_list(values)

        if sort:
            values = sorted(values)

        return values

    def _get_requirements(
        self,
        extras: Iterable[str],
        groups: Iterable[str],
        skip_package: bool = False,
    ) -> list[Requirement]:
        out: list[Requirement] = []

        if not skip_package:
            out.extend(self.requirements_base)

        def _extend_extra_or_group(
            extras: Iterable[str],
            requirements_mapping: dict[str, list[Requirement]],
        ) -> None:
            for extra in extras:
                out.extend(requirements_mapping[extra])

        _extend_extra_or_group(extras, self.requirements_extras)
        _extend_extra_or_group(groups, self.requirements_groups)

        return out

    def pip_requirements(
        self,
        *,
        extras: str | Iterable[str] | None = None,
        groups: str | Iterable[str] | None = None,
        extras_or_groups: str | Iterable[str] | None = None,
        skip_package: bool = False,
        pip_deps: str | Iterable[str] | None = None,
        unique: bool = True,
        remove_whitespace: bool = True,
        sort: bool = True,
    ) -> list[str]:
        """Pip dependencies."""
        extras, groups = self._resolve_extras_and_groups(
            extras, groups, extras_or_groups
        )

        out: list[str] = [
            str(requirement)
            for requirement in self._get_requirements(
                extras=extras,
                groups=groups,
                skip_package=skip_package,
            )
        ]

        if pip_deps:
            pip_deps = [pip_deps] if isinstance(pip_deps, str) else list(pip_deps)
            out.extend(_clean_pip_reqs(pip_deps))

        return self._cleanup(
            out, remove_whitespace=remove_whitespace, unique=unique, sort=sort
        )

    def conda_and_pip_requirements(  # noqa: C901
        self,
        *,
        extras: str | Iterable[str] | None = None,
        groups: str | Iterable[str] | None = None,
        extras_or_groups: str | Iterable[str] | None = None,
        skip_package: bool = False,
        pip_deps: str | Iterable[str] | None = None,
        conda_deps: str | Iterable[str] | None = None,
        unique: bool = True,
        remove_whitespace: bool = True,
        sort: bool = True,
        python_version: str | None = None,
        python_include: str | None = None,
    ) -> tuple[list[str], list[str]]:
        """Conda and pip requirements."""

        def _init_deps(deps: str | Iterable[str] | None) -> list[str]:
            if deps is None:
                return []
            if isinstance(deps, str):
                return [deps]
            return list(deps)

        extras, groups = self._resolve_extras_and_groups(
            extras, groups, extras_or_groups
        )

        if python_include == "infer":
            # safer get
            if (x := self.get_in("project", "requires-python")) is None:
                msg = "No value for `requires-python` in pyproject.toml file"
                raise ValueError(msg)
            python_include = "python" + str(x)

        pip_deps = _clean_pip_reqs(_init_deps(pip_deps))
        conda_deps = _clean_conda_strings(
            _init_deps(conda_deps), python_version=python_version
        )

        override_table = self.override_table
        for requirement in self._get_requirements(
            extras=extras,
            groups=groups,
            skip_package=skip_package,
        ):
            override = override_table.get(requirement.name)
            if override is not None:
                if override.pip:
                    pip_deps.append(str(requirement))

                elif not override.skip:
                    r = _clean_conda_requirement(
                        requirement,
                        python_version=python_version,
                        channel=override.channel,
                    )

                    if r is not None:
                        conda_deps.append(str(r))

                conda_deps.extend(
                    _clean_conda_strings(
                        override.packages, python_version=python_version
                    )
                )
            else:
                r = _clean_conda_requirement(requirement, python_version=python_version)
                if r is not None:  # pragma: no cover
                    conda_deps.append(str(r))

        pip_deps, conda_deps = (
            self._cleanup(
                x, remove_whitespace=remove_whitespace, unique=unique, sort=sort
            )
            for x in (pip_deps, conda_deps)
        )

        if python_include is not None:
            conda_deps = [
                *self._cleanup([python_include], remove_whitespace=remove_whitespace),
                *conda_deps,
            ]

        # special if have pip requirements or just pip in conda_deps
        # in this case, make sure pip is last
        if "pip" in conda_deps:
            conda_deps.remove("pip")
            conda_deps.append("pip")
        elif pip_deps:
            conda_deps.append("pip")

        return conda_deps, pip_deps

    def to_conda_yaml(  # noqa: PLR0913
        self,
        *,
        extras: str | Iterable[str] | None = None,
        groups: str | Iterable[str] | None = None,
        extras_or_groups: str | Iterable[str] | None = None,
        pip_deps: str | Iterable[str] | None = None,
        conda_deps: str | Iterable[str] | None = None,
        name: str | None = None,
        channels: str | Iterable[str] | None = None,
        python_include: str | None = None,
        python_version: str | None = None,
        skip_package: bool = False,
        header_cmd: str | None = None,
        output: str | Path | None = None,
        sort: bool = True,
        remove_whitespace: bool = True,
        unique: bool = True,
        allow_empty: bool = False,
    ) -> str:
        """Create yaml string."""
        conda_deps, pip_deps = self.conda_and_pip_requirements(
            extras=extras,
            groups=groups,
            extras_or_groups=extras_or_groups,
            skip_package=skip_package,
            pip_deps=pip_deps,
            conda_deps=conda_deps,
            unique=unique,
            remove_whitespace=remove_whitespace,
            sort=sort,
            python_include=python_include,
            python_version=python_version,
        )

        if not conda_deps and not pip_deps:
            return _check_allow_empty(allow_empty)

        out = _conda_yaml(
            name=name,
            channels=channels or self.channels,
            conda_deps=conda_deps,
            pip_deps=pip_deps,
        )

        out = _add_header(out, header_cmd)

        _optional_write(out, output)

        return out

    def to_requirements(
        self,
        *,
        extras: str | Iterable[str] | None = None,
        groups: str | Iterable[str] | None = None,
        extras_or_groups: str | Iterable[str] | None = None,
        skip_package: bool = False,
        header_cmd: str | None = None,
        output: str | Path | None = None,
        sort: bool = True,
        pip_deps: Sequence[str] | None = None,
        allow_empty: bool = False,
        remove_whitespace: bool = True,
    ) -> str:
        """Create requirements string."""
        pip_deps = self.pip_requirements(
            extras=extras,
            groups=groups,
            extras_or_groups=extras_or_groups,
            skip_package=skip_package,
            pip_deps=pip_deps,
            remove_whitespace=remove_whitespace,
            sort=sort,
        )

        if not pip_deps:
            return _check_allow_empty(allow_empty)

        out = _add_header(list_to_str(pip_deps), header_cmd)

        _optional_write(out, output)
        return out

    def to_conda_requirements(  # noqa: PLR0913
        self,
        *,
        extras: str | Iterable[str] | None = None,
        groups: str | Iterable[str] | None = None,
        extras_or_groups: str | Iterable[str] | None = None,
        channels: str | Iterable[str] | None = None,
        python_include: str | None = None,
        python_version: str | None = None,
        prepend_channel: bool = False,
        output_conda: str | Path | None = None,
        output_pip: str | Path | None = None,
        skip_package: bool = False,
        header_cmd: str | None = None,
        sort: bool = True,
        unique: bool = True,
        conda_deps: str | Iterable[str] | None = None,
        pip_deps: str | Iterable[str] | None = None,
        remove_whitespace: bool = True,
    ) -> tuple[str, str]:
        """Create conda and pip requirements files."""
        conda_deps, pip_deps = self.conda_and_pip_requirements(
            extras=extras,
            groups=groups,
            extras_or_groups=extras_or_groups,
            skip_package=skip_package,
            pip_deps=pip_deps,
            conda_deps=conda_deps,
            unique=unique,
            remove_whitespace=remove_whitespace,
            sort=sort,
            python_include=python_include,
            python_version=python_version,
        )

        if channels:
            channels = (
                [channels] if isinstance(channels, str) else list(channels)
            )  # pragma: no cover
        else:
            channels = self.channels

        if conda_deps and channels and prepend_channel:
            if len(channels) != 1:
                msg = "Can only pass single channel to prepend."
                raise ValueError(msg)
            channel = channels[0]
            # add in channel if none exists
            conda_deps = [
                dep if "::" in dep else f"{channel}::{dep}" for dep in conda_deps
            ]

        conda_deps_str = _add_header(list_to_str(conda_deps), header_cmd)
        pip_deps_str = _add_header(list_to_str(pip_deps), header_cmd)

        if output_conda and conda_deps_str:
            _optional_write(conda_deps_str, output_conda)

        if output_pip and pip_deps_str:
            _optional_write(pip_deps_str, output_pip)

        return conda_deps_str, pip_deps_str

    @classmethod
    def from_string(
        cls,
        toml_string: str,
    ) -> Self:
        """Create object from string."""
        from ._compat import tomllib

        data = tomllib.loads(toml_string)
        return cls(data=data)

    @classmethod
    def from_path(cls, path: str | Path) -> Self:
        """Create object from path."""
        from ._compat import tomllib

        with Path(path).open("rb") as f:
            data = tomllib.load(f)
        return cls(data=data)



# /// script
# requires-python = ">=3.11"
# dependencies = [
#     "jupyter",
# ]
# ///
"""
Script to cleanup installed ipykernels that no longer exist.

This script must be run from the environment with the notebook server.
For example, if you have a conda environment `notebook` with your
jupyter server, run the following:

$ conda run -n notebook python path/to/clean_kernelspec.py
"""

from __future__ import annotations

import json
import logging
from pathlib import Path
from subprocess import CalledProcessError

FORMAT = "[%(name)s - %(levelname)s] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger("clean_kernelspec")


def get_kernelspec_data() -> None:
    """Main function."""
    from subprocess import check_output

    s = check_output(
        ["jupyter", "kernelspec", "list", "--json", "--log-level", "ERROR"],
    )

    to_remove: list[str] = []
    for name, data in json.loads(s)["kernelspecs"].items():
        p = Path(data["spec"]["argv"][0])

        if not p.exists():
            logger.debug("%s does not exist.", name)
            to_remove.append(name)

        else:
            logger.debug("%s exists.", name)

    if to_remove:
        logger.info("removing kernels %s", to_remove)
        check_output(["jupyter", "kernelspec", "remove", "-f", *to_remove])
    else:
        logger.info("nothing to do")


if __name__ == "__main__":
    try:
        get_kernelspec_data()
    except CalledProcessError:
        logger.exception("Most likely you didn't run from notebook server environment")
        raise



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, TypeVar

R = TypeVar("R")
T = TypeVar("T")


if TYPE_CHECKING:
    from typing import Literal, Tuple

    from packaging.requirements import Requirement

    from .overrides import OverrideDeps
    from .utils import _Missing  # pyright: ignore[reportPrivateUsage]

    MISSING_TYPE = Literal[_Missing.MISSING]

    RequirementCommentPair = Tuple[Optional[Requirement], Optional[str]]
    RequirementOverridePair = Tuple[Optional[Requirement], Optional[OverrideDeps]]



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
from __future__ import annotations

import locale
import tempfile
from textwrap import dedent

import pytest

from pyproject2conda import requirements
from pyproject2conda.utils import get_in


def test_get_in() -> None:
    d = {"a": {"b": {"c": 3}}}

    assert get_in(["a", "b", "c"], d) == 3

    assert get_in(["a", "d"], d, "hello") == "hello"


def test_list_to_string() -> None:
    from pyproject2conda.utils import list_to_str

    assert list_to_str(["a", "b"], eol=True) == "a\nb\n"
    assert list_to_str(["a", "b"], eol=False) == "a\nb"
    assert list_to_str(None) == ""  # noqa: PLC1901


def test_header() -> None:
    from pyproject2conda.requirements import _create_header

    expected = dedent(
        """\
#
# This file is autogenerated by pyproject2conda.
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#"""
    )

    assert expected == _create_header()

    cmd = "hello"
    out = _create_header(cmd=cmd)

    header = dedent(
        f"""\
#
# This file is autogenerated by pyproject2conda
# with the following command:
#
#     $ {cmd}
#
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#"""
    )

    assert out == header


def test_optional_write() -> None:
    from pathlib import Path

    from pyproject2conda.requirements import _optional_write

    s = "hello"
    with tempfile.TemporaryDirectory() as d:
        p = Path(d) / "tmp.txt"

        _optional_write(s, p)

        with Path(p).open(encoding=locale.getpreferredencoding(False)) as f:
            test = f.read()

        assert test == s


def test_output_to_yaml() -> None:
    from pyproject2conda.requirements import _conda_yaml

    with pytest.raises(ValueError):
        s = _conda_yaml()

    s = _conda_yaml(
        conda_deps=["a", "pip"],
        channels=["conda-forge"],
        pip_deps=["pip-thing"],
        name="hello",
    )

    expected = """\
name: hello
channels:
  - conda-forge
dependencies:
  - a
  - pip
  - pip:
      - pip-thing
    """

    assert s == dedent(expected)


def test_infer() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    dependencies = [
    "athing",
    "bthing",
    "cthing; python_version<'3.10'",
    ]

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest",
    ]
    dev-extras = [
    "matplotlib",
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]
    dist-pypi = [
    "setuptools",
    "build",
    ]


    [tool.pyproject2conda]
    channels = 'conda-forge'

    [tool.pyproject2conda.dependencies]
    athing = { pip = true }
    bthing = { skip = true, packages = "bthing-conda" }
    cthing = { channel = "conda-forge" }
    pytest = { channel = "conda-forge" }
    matplotlib = { skip = true, packages = [
        "additional-thing",
        "conda-matplotlib"
    ] }

    """
    )

    d = requirements.ParseDepends.from_string(toml)
    with pytest.raises(ValueError):
        d.to_conda_yaml(python_include="infer")


def test_channels() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    """
    )

    d = requirements.ParseDepends.from_string(toml)

    assert d.channels == []

    toml = dedent(
        """\
    [project]
    name = "hello"

    [tool.pyproject2conda]
    channels = 'conda-forge'
    """
    )

    d = requirements.ParseDepends.from_string(toml)

    assert d.channels == ["conda-forge"]


def test_pip_requirements() -> None:
    toml = dedent(
        """\
    [project]
    name="hello"
    requires-python = ">=3.8,<3.11"
    dependencies = [
    "athing",
    "bthing",
    "cthing; python_version<'3.10'",
    ]

    [project.optional-dependencies]
    test = ["pytest", "test-optional"]
    dev = ["hello[test]", "dev-package"]

    [dependency-groups]
    test = ["pytest", "test-optional-group"]
    dev = [{include-group= "test"}, "dev-package-group"]
        """
    )

    expected = dedent(
        """\
    athing
    bthing
    cthing;python_version<"3.10"
    hello
    """
    )

    d = requirements.ParseDepends.from_string(toml)

    assert d.to_requirements(pip_deps="hello") == expected

    expected = dedent(
        """\
        athing
        bthing
        cthing;python_version<"3.10"
        dev-package
        hello
        pytest
        test-optional
        """
    )

    d = requirements.ParseDepends.from_string(toml)
    assert d.to_requirements(pip_deps="hello", extras="dev") == expected
    assert d.to_requirements(pip_deps="hello", extras_or_groups="dev") == expected

    expected = dedent(
        """\
        athing
        bthing
        cthing;python_version<"3.10"
        dev-package-group
        hello
        pytest
        test-optional-group
        """
    )

    d = requirements.ParseDepends.from_string(toml)
    assert d.to_requirements(pip_deps="hello", groups="dev") == expected

    # build-system.requires


def test_to_conda_requirements_error() -> None:
    toml = dedent(
        """\
    [project]
    requires-python = ">=3.8,<3.11"
    dependencies = [
    "athing",
    "bthing",
    "cthing; python_version<'3.10'",
    ]

    [tool.pyproject2conda.dependencies]
    athing = { pip = true }
    bthing = { skip = true, packages = "bthing-conda" }
    cthing = { channel = "conda-forge" }
        """
    )

    d = requirements.ParseDepends.from_string(toml)

    with pytest.raises(ValueError):
        d.to_conda_requirements(channels=["hello", "there"], prepend_channel=True)


def test_package_name() -> None:
    toml = dedent(
        """\
    [project]
    requires-python = ">=3.8,<3.11"
    dependencies = [
    "athing",
    "bthing",
    "cthing; python_version<'3.10'",
    ]

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest",
    ]
    dev-extras = [
    "matplotlib",
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]
    dist-pypi = [
    "setuptools",
    "build",
    ]


    [tool.pyproject2conda]
    channels = ['conda-forge']

    [tool.pyproject2conda.dependencies]
    athing = { pip = true }
    bthing = { skip = true, packages = "bthing-conda" }
    cthing = { channel = "conda-forge" }
    pytest = { channel = "conda-forge" }
    matplotlib = { skip = true, packages = [
        "additional-thing",
        "conda-matplotlib"
    ] }
    """
    )
    d = requirements.ParseDepends.from_string(toml)
    with pytest.raises(ValueError):
        d.conda_and_pip_requirements(extras="dev")


def test_conda_and_pip_simple() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    dependencies = [
    "a",
    # thing
    "b",
    ]
        """
    )

    d = requirements.ParseDepends.from_string(toml)

    conda_deps, pip_deps = d.conda_and_pip_requirements()

    assert conda_deps == ["a", "b"]

    assert pip_deps == []


@pytest.mark.parametrize(
    ("style", "toml"),
    [
        # Extras
        (
            "extras",
            dedent(
                """\
                [build-system]
                requires = ["setuptools>=61.2", "setuptools_scm[toml]>=8.0"]
                build-backend = "setuptools.build_meta"

                [project]
                name = "hello"
                requires-python = ">=3.8, <3.11"
                dependencies = [
                "athing",
                "bthing",
                "cthing; python_version<'3.10'",
                ]

                [project.optional-dependencies]
                test = [
                "pandas",
                "pytest",
                ]
                dev-extras = [
                "matplotlib",
                ]
                dev = [
                "hello[test]",
                "hello[dev-extras]",
                ]
                dist-pypi = [
                "setuptools",
                "build",
                ]


                [tool.pyproject2conda]
                channels = ['conda-forge']

                [tool.pyproject2conda.dependencies]
                athing = {pip = true}
                bthing = {skip = true, packages = "bthing-conda"}
                cthing = {channel = "conda-forge"}
                pytest = {channel = "conda-forge"}
                matplotlib = {skip = true, packages = ["additional-thing", "conda-matplotlib"]}
                build = { channel = "pip" }
                """
            ),
        ),
        # Groups
        (
            "groups",
            dedent(
                """\
                [build-system]
                requires = ["setuptools>=61.2", "setuptools_scm[toml]>=8.0"]
                build-backend = "setuptools.build_meta"

                [project]
                name = "hello"
                requires-python = ">=3.8, <3.11"
                dependencies = [
                "athing",
                "bthing",
                "cthing; python_version<'3.10'",
                ]

                [dependency-groups]
                test = [
                "pandas",
                "pytest",
                ]
                dev-extras = [
                "matplotlib",
                ]
                dev = [
                {include-group = "test"},
                {include-group = "dev-extras"},
                ]
                dist-pypi = [
                "setuptools",
                "build",
                ]


                [tool.pyproject2conda]
                channels = ['conda-forge']

                [tool.pyproject2conda.dependencies]
                athing = {pip = true}
                bthing = {skip = true, packages = "bthing-conda"}
                cthing = {channel = "conda-forge"}
                pytest = {channel = "conda-forge"}
                matplotlib = {skip = true, packages = ["additional-thing", "conda-matplotlib"]}
                build = { channel = "pip" }
                """
            ),
        ),
    ],
)
def test_complete(style, toml) -> None:
    d = requirements.ParseDepends.from_string(toml)

    # test unknown extra
    with pytest.raises(ValueError):
        d.to_conda_yaml(extras="a-thing")

    # test list:
    assert getattr(d, style) == [
        "test",
        "dev-extras",
        "dev",
        "dist-pypi",
        "build-system.requires",
    ]

    # test build-system.requires
    expected = """\
    setuptools>=61.2
    setuptools_scm[toml]>=8.0
    """
    assert dedent(expected) == d.to_requirements(
        skip_package=True,
        **{style: "build-system.requires"},  # type: ignore[arg-type]
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml()

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - -e.
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml(pip_deps="-e .")

    # test -p option
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(python_include="infer")

    # with remove spaces:
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8, <3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=False
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.9
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(python_include="python=3.9")

    # test passing python_version
    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.10
  - bthing-conda
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="python=3.10", python_version="3.10"
    )

    out = d.to_conda_yaml(channels="hello")

    expected = """\
channels:
  - hello
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(**{style: "test"}, sort=False)  # type: ignore[arg-type]

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(**{style: "test"}, sort=True)  # type: ignore[arg-type]

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(**{style: "dist-pypi"}, skip_package=True)  # type: ignore[arg-type]

    expected = """\
channels:
  - conda-forge
dependencies:
  - setuptools
  - pip
  - pip:
      - build
    """

    assert out == dedent(expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - additional-thing
  - conda-matplotlib
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml(**{style: "dev"}, sort=False)  # type: ignore[arg-type]

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml(**{style: "dev"})  # type: ignore[arg-type]

    # Test deps/reqs
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - dep
  - pip
  - pip:
      - athing
      - req
    """

    assert dedent(expected) == d.to_conda_yaml(
        conda_deps="dep;python_version<'3.10'", pip_deps=["req"]
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.10
  - bthing-conda
  - pip
  - pip:
      - athing
      - req;python_version<"3.10"
    """
    assert dedent(expected) == d.to_conda_yaml(
        conda_deps=["dep;python_version<'3.10'"],
        pip_deps=["req;python_version<'3.10'"],
        python_include="python=3.10",
        python_version="3.10",
    )


def test_missing_dependencies() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"
    dependencies = []

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest",
    "pytest-accept",
    ]
    dev-extras = [

    "matplotlib",
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']

    [tool.pyproject2conda.dependencies]
    pytest = { channel = "conda-forge" }
    pytest-accept = { pip = true }
    matplotlib = { skip = true, packages = [
        "additional-thing",
        "conda-matplotlib"
    ] }
    """
    )

    d = requirements.ParseDepends.from_string(toml)

    expected = """\
channels:
  - conda-forge
dependencies:
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - pytest-accept
    """

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        skip_package=True,
    )

    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest",
    "pytest-accept",
    ]
    dev-extras = [

    "matplotlib",
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']

    [tool.pyproject2conda.dependencies]
    pytest = { channel = "conda-forge" }
    pytest-accept = { pip = true }
    matplotlib = { skip = true, packages = [
        "additional-thing",
        "conda-matplotlib"
    ] }
    """
    )

    d = requirements.ParseDepends.from_string(toml)

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        skip_package=True,
    )

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        skip_package=False,
    )

    # check output has no dependencies:
    for attr in ["to_conda_yaml", "to_requirements"]:
        f = getattr(d, attr)
        with pytest.raises(ValueError):
            f()

        assert f(allow_empty=True) == "No dependencies for this environment\n"


def test_clean_conda() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest",
    "pytest-accept",
    ]
    dev-extras = [

    "matplotlib",
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']


    [tool.pyproject2conda.dependencies]
    pytest = { channel = "conda-forge" }
    pytest-accept = { pip = true }
    matplotlib = { skip = true, packages = [
        "additional-thing",
        "conda-forge::conda-matplotlib"
    ] }
    """
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - conda-forge::conda-matplotlib
    """

    d = requirements.ParseDepends.from_string(toml)

    assert dedent(expected) == d.to_conda_yaml(
        extras="dev-extras",
        skip_package=True,
    )


def test_no_optional_deps() -> None:
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"
    dependencies = [
    "athing >0.5",
    "bthing > 1.0",
    "cthing; python_version < '3.10'",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']

    [tool.pyproject2conda.dependencies]
    athing = { pip = true }
    bthing = { skip = true, packages = "bthing-conda > 2.0" }
    cthing = { channel = "conda-forge" }
    """
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda>2.0
  - conda-forge::cthing
  - pip
  - pip:
      - athing>0.5
    """

    d = requirements.ParseDepends.from_string(toml)

    assert d.optional_dependencies == {}

    assert dedent(expected) == d.to_conda_yaml(python_include="infer")


@pytest.mark.parametrize(
    "toml",
    [
        dedent(
            """\
            [project]
            name = "hello"
            requires-python = ">=3.8, <3.11"
            dependencies = [
            "athing >0.5",
            "bthing > 1.0",
            "cthing; python_version < '3.10'",
            ]


            [tool.pyproject2conda]
            channels = ['conda-forge']

            [tool.pyproject2conda.dependencies]
            athing = {pip = true}
            bthing = {skip = true, packages = "bthing-conda > 2.0"}
            cthing = {channel = "conda-forge"}
            """
        ),
    ],
)
def test_spaces(toml) -> None:
    d = requirements.ParseDepends.from_string(toml)

    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8, <3.11
  - bthing-conda>2.0
  - conda-forge::cthing
  - pip
  - pip:
      - athing>0.5
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=False
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda>2.0
  - conda-forge::cthing
  - pip
  - pip:
      - athing>0.5
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=True
    )

    expected = """\
    athing>0.5
    bthing>1.0
    cthing; python_version < "3.10"
    """

    assert dedent(expected) == d.to_requirements(remove_whitespace=False)

    expected = """\
    athing>0.5
    bthing>1.0
    cthing;python_version<"3.10"
    """

    assert dedent(expected) == d.to_requirements(remove_whitespace=True)


def test_include_pip() -> None:
    toml = dedent(
        """\
        [build-system]
        requires = ["setuptools>=61.2", "setuptools_scm[toml]>=8.0"]
        build-backend = "setuptools.build_meta"

        [project]
        name = "hello"
        requires-python = ">=3.8, <3.11"
        dependencies = [
        "athing",
        ]

        [project.optional-dependencies]
        test = [
        "xthing",
        ]
        """
    )

    expected = """\
    dependencies:
      - athing
      - xthing
      - pip
    """

    d = requirements.ParseDepends.from_string(toml)

    assert dedent(expected) == d.to_conda_yaml(extras="test", conda_deps="pip")
